"""
Smart contract ABIs and interfaces for ERC-8004.
"""

from typing import Dict, List, Any

# ERC-721 ABI (minimal required functions)
ERC721_ABI = [
    {
        "inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
        "name": "ownerOf",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "operator", "type": "address"}
        ],
        "name": "isApprovedForAll",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
        "name": "getApproved",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "tokenId", "type": "uint256"}
        ],
        "name": "transferFrom",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "bool", "name": "approved", "type": "bool"}
        ],
        "name": "setApprovalForAll",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
            {"internalType": "address", "name": "to", "type": "address"}
        ],
        "name": "approve",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

# ERC-721 URI Storage ABI
ERC721_URI_STORAGE_ABI = [
    {
        "inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
        "name": "tokenURI",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
            {"internalType": "string", "name": "_tokenURI", "type": "string"}
        ],
        "name": "setTokenURI",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    }
]

# Identity Registry ABI
IDENTITY_REGISTRY_ABI = [
    # ERC-721 functions
    *ERC721_ABI,
    *ERC721_URI_STORAGE_ABI,
    
    # Identity Registry specific functions
    {
        "inputs": [],
        "name": "register",
        "outputs": [{"internalType": "uint256", "name": "agentId", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "string", "name": "tokenUri", "type": "string"}],
        "name": "register",
        "outputs": [{"internalType": "uint256", "name": "agentId", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "string", "name": "tokenUri", "type": "string"},
            {
                "components": [
                    {"internalType": "string", "name": "key", "type": "string"},
                    {"internalType": "bytes", "name": "value", "type": "bytes"}
                ],
                "internalType": "struct IdentityRegistry.MetadataEntry[]",
                "name": "metadata",
                "type": "tuple[]"
            }
        ],
        "name": "register",
        "outputs": [{"internalType": "uint256", "name": "agentId", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "string", "name": "key", "type": "string"}
        ],
        "name": "getMetadata",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "string", "name": "key", "type": "string"},
            {"internalType": "bytes", "name": "value", "type": "bytes"}
        ],
        "name": "setMetadata",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "string", "name": "newUri", "type": "string"}
        ],
        "name": "setAgentUri",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    
    # Events
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": False, "internalType": "string", "name": "tokenURI", "type": "string"},
            {"indexed": True, "internalType": "address", "name": "owner", "type": "address"}
        ],
        "name": "Registered",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": True, "internalType": "string", "name": "indexedKey", "type": "string"},
            {"indexed": False, "internalType": "string", "name": "key", "type": "string"},
            {"indexed": False, "internalType": "bytes", "name": "value", "type": "bytes"}
        ],
        "name": "MetadataSet",
        "type": "event"
    }
]

# Reputation Registry ABI
REPUTATION_REGISTRY_ABI = [
    {
        "inputs": [],
        "name": "getIdentityRegistry",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "uint8", "name": "score", "type": "uint8"},
            {"internalType": "bytes32", "name": "tag1", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag2", "type": "bytes32"},
            {"internalType": "string", "name": "feedbackUri", "type": "string"},
            {"internalType": "bytes32", "name": "feedbackHash", "type": "bytes32"},
            {"internalType": "bytes", "name": "feedbackAuth", "type": "bytes"}
        ],
        "name": "giveFeedback",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "uint64", "name": "feedbackIndex", "type": "uint64"}
        ],
        "name": "revokeFeedback",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address", "name": "clientAddress", "type": "address"},
            {"internalType": "uint64", "name": "feedbackIndex", "type": "uint64"},
            {"internalType": "string", "name": "responseUri", "type": "string"},
            {"internalType": "bytes32", "name": "responseHash", "type": "bytes32"}
        ],
        "name": "appendResponse",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address", "name": "clientAddress", "type": "address"}
        ],
        "name": "getLastIndex",
        "outputs": [{"internalType": "uint64", "name": "", "type": "uint64"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address", "name": "clientAddress", "type": "address"},
            {"internalType": "uint64", "name": "index", "type": "uint64"}
        ],
        "name": "readFeedback",
        "outputs": [
            {"internalType": "uint8", "name": "score", "type": "uint8"},
            {"internalType": "bytes32", "name": "tag1", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag2", "type": "bytes32"},
            {"internalType": "bool", "name": "isRevoked", "type": "bool"}
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address[]", "name": "clientAddresses", "type": "address[]"},
            {"internalType": "bytes32", "name": "tag1", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag2", "type": "bytes32"}
        ],
        "name": "getSummary",
        "outputs": [
            {"internalType": "uint64", "name": "count", "type": "uint64"},
            {"internalType": "uint8", "name": "averageScore", "type": "uint8"}
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address[]", "name": "clientAddresses", "type": "address[]"},
            {"internalType": "bytes32", "name": "tag1", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag2", "type": "bytes32"},
            {"internalType": "bool", "name": "includeRevoked", "type": "bool"}
        ],
        "name": "readAllFeedback",
        "outputs": [
            {"internalType": "address[]", "name": "clients", "type": "address[]"},
            {"internalType": "uint8[]", "name": "scores", "type": "uint8[]"},
            {"internalType": "bytes32[]", "name": "tag1s", "type": "bytes32[]"},
            {"internalType": "bytes32[]", "name": "tag2s", "type": "bytes32[]"},
            {"internalType": "bool[]", "name": "revokedStatuses", "type": "bool[]"}
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "uint256", "name": "agentId", "type": "uint256"}],
        "name": "getClients",
        "outputs": [{"internalType": "address[]", "name": "", "type": "address[]"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address", "name": "clientAddress", "type": "address"},
            {"internalType": "uint64", "name": "feedbackIndex", "type": "uint64"},
            {"internalType": "address[]", "name": "responders", "type": "address[]"}
        ],
        "name": "getResponseCount",
        "outputs": [{"internalType": "uint64", "name": "count", "type": "uint64"}],
        "stateMutability": "view",
        "type": "function"
    },
    
    # Events
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": True, "internalType": "address", "name": "clientAddress", "type": "address"},
            {"indexed": False, "internalType": "uint8", "name": "score", "type": "uint8"},
            {"indexed": True, "internalType": "bytes32", "name": "tag1", "type": "bytes32"},
            {"indexed": False, "internalType": "bytes32", "name": "tag2", "type": "bytes32"},
            {"indexed": False, "internalType": "string", "name": "feedbackUri", "type": "string"},
            {"indexed": False, "internalType": "bytes32", "name": "feedbackHash", "type": "bytes32"}
        ],
        "name": "NewFeedback",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": True, "internalType": "address", "name": "clientAddress", "type": "address"},
            {"indexed": True, "internalType": "uint64", "name": "feedbackIndex", "type": "uint64"}
        ],
        "name": "FeedbackRevoked",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": True, "internalType": "address", "name": "clientAddress", "type": "address"},
            {"indexed": False, "internalType": "uint64", "name": "feedbackIndex", "type": "uint64"},
            {"indexed": True, "internalType": "address", "name": "responder", "type": "address"},
            {"indexed": False, "internalType": "string", "name": "responseUri", "type": "string"},
            {"indexed": False, "internalType": "bytes32", "name": "responseHash", "type": "bytes32"}
        ],
        "name": "ResponseAppended",
        "type": "event"
    }
]

# Validation Registry ABI
VALIDATION_REGISTRY_ABI = [
    {
        "inputs": [],
        "name": "getIdentityRegistry",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "address", "name": "validatorAddress", "type": "address"},
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "string", "name": "requestUri", "type": "string"},
            {"internalType": "bytes32", "name": "requestHash", "type": "bytes32"}
        ],
        "name": "validationRequest",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "requestHash", "type": "bytes32"},
            {"internalType": "uint8", "name": "response", "type": "uint8"},
            {"internalType": "string", "name": "responseUri", "type": "string"},
            {"internalType": "bytes32", "name": "responseHash", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag", "type": "bytes32"}
        ],
        "name": "validationResponse",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "requestHash", "type": "bytes32"}],
        "name": "getValidationStatus",
        "outputs": [
            {"internalType": "address", "name": "validatorAddress", "type": "address"},
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "uint8", "name": "response", "type": "uint8"},
            {"internalType": "bytes32", "name": "responseHash", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag", "type": "bytes32"},
            {"internalType": "uint256", "name": "lastUpdate", "type": "uint256"}
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "address[]", "name": "validatorAddresses", "type": "address[]"},
            {"internalType": "bytes32", "name": "tag", "type": "bytes32"}
        ],
        "name": "getSummary",
        "outputs": [
            {"internalType": "uint64", "name": "count", "type": "uint64"},
            {"internalType": "uint8", "name": "avgResponse", "type": "uint8"}
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "uint256", "name": "agentId", "type": "uint256"}],
        "name": "getAgentValidations",
        "outputs": [{"internalType": "bytes32[]", "name": "", "type": "bytes32[]"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "address", "name": "validatorAddress", "type": "address"}],
        "name": "getValidatorRequests",
        "outputs": [{"internalType": "bytes32[]", "name": "", "type": "bytes32[]"}],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "validations",
        "outputs": [
            {"internalType": "address", "name": "validatorAddress", "type": "address"},
            {"internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"internalType": "uint8", "name": "response", "type": "uint8"},
            {"internalType": "bytes32", "name": "responseHash", "type": "bytes32"},
            {"internalType": "bytes32", "name": "tag", "type": "bytes32"},
            {"internalType": "uint256", "name": "lastUpdate", "type": "uint256"}
        ],
        "stateMutability": "view",
        "type": "function"
    },
    
    # Events
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "address", "name": "validatorAddress", "type": "address"},
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": False, "internalType": "string", "name": "requestUri", "type": "string"},
            {"indexed": True, "internalType": "bytes32", "name": "requestHash", "type": "bytes32"}
        ],
        "name": "ValidationRequest",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {"indexed": True, "internalType": "address", "name": "validatorAddress", "type": "address"},
            {"indexed": True, "internalType": "uint256", "name": "agentId", "type": "uint256"},
            {"indexed": True, "internalType": "bytes32", "name": "requestHash", "type": "bytes32"},
            {"indexed": False, "internalType": "uint8", "name": "response", "type": "uint8"},
            {"indexed": False, "internalType": "string", "name": "responseUri", "type": "string"},
            {"indexed": False, "internalType": "bytes32", "name": "responseHash", "type": "bytes32"},
            {"indexed": False, "internalType": "bytes32", "name": "tag", "type": "bytes32"}
        ],
        "name": "ValidationResponse",
        "type": "event"
    }
]

# Contract registry for different chains
DEFAULT_REGISTRIES: Dict[int, Dict[str, str]] = {
    11155111: {  # Ethereum Sepolia
        "IDENTITY": "0x8004a6090Cd10A7288092483047B097295Fb8847",
        "REPUTATION": "0x8004B8FD1A363aa02fDC07635C0c5F94f6Af5B7E",
        "VALIDATION": "0x8004CB39f29c09145F24Ad9dDe2A108C1A2cdfC5",
    },
    84532: {  # Base Sepolia
        "IDENTITY": "0x8004AA63c570c570eBF15376c0dB199918BFe9Fb",
        "REPUTATION": "0x8004bd8daB57f14Ed299135749a5CB5c42d341BF",
        "VALIDATION": "0x8004C269D0A5647E51E121FeB226200ECE932d55",
    },
    80002: {  # Polygon Amoy
        "IDENTITY": "0x8004ad19E14B9e0654f73353e8a0B600D46C2898",
        "REPUTATION": "0x8004B12F4C2B42d00c46479e859C92e39044C930",
        "VALIDATION": "0x8004C11C213ff7BaD36489bcBDF947ba5eee289B",
    },
    59141: {  # Linea Sepolia
        "IDENTITY": "0x8004aa7C931bCE1233973a0C6A667f73F66282e7",
        "REPUTATION": "0x8004bd8483b99310df121c46ED8858616b2Bba02",
        "VALIDATION": "0x8004c44d1EFdd699B2A26e781eF7F77c56A9a4EB",
    },
}

# Default subgraph URLs for different chains
DEFAULT_SUBGRAPH_URLS: Dict[int, str] = {
    11155111: "https://gateway.thegraph.com/api/00a452ad3cd1900273ea62c1bf283f93/subgraphs/id/6wQRC7geo9XYAhckfmfo8kbMRLeWU8KQd3XsJqFKmZLT",  # Ethereum Sepolia
    84532: "https://gateway.thegraph.com/api/00a452ad3cd1900273ea62c1bf283f93/subgraphs/id/GjQEDgEKqoh5Yc8MUgxoQoRATEJdEiH7HbocfR1aFiHa",  # Base Sepolia
    80002: "https://gateway.thegraph.com/api/00a452ad3cd1900273ea62c1bf283f93/subgraphs/id/2A1JB18r1mF2VNP4QBH4mmxd74kbHoM6xLXC8ABAKf7j",  # Polygon Amoy
}



"""
Test for Agent Feedback Flow with IPFS Pin
Submits feedback from a client to an existing agent and verifies data integrity.

Flow:
1. Load existing agent by ID
2. Client submits multiple feedback entries
3. Verify feedback data consistency (score, tags, capability, skill)
4. Wait for blockchain finalization
5. Verify feedback can be retrieved (if SDK supports it)

Usage:
    Update AGENT_ID constant below to point to your existing agent
"""

import logging
import time
import random
import sys

# Configure logging: root logger at WARNING to suppress noisy dependencies
logging.basicConfig(
    level=logging.WARNING,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    handlers=[logging.StreamHandler(sys.stdout)]
)

# Set debug level ONLY for agent0_sdk
logging.getLogger('agent0_sdk').setLevel(logging.DEBUG)
logging.getLogger('agent0_sdk.core').setLevel(logging.DEBUG)

from agent0_sdk import SDK
from config import CHAIN_ID, RPC_URL, AGENT_PRIVATE_KEY, PINATA_JWT, SUBGRAPH_URL, AGENT_ID, print_config

# Client configuration (different wallet)
CLIENT_PRIVATE_KEY = "f8d368064ccf80769e348a59155f69ec224849bd507a8c26dd85beefa777331a"


def generateFeedbackData(index: int):
    """Generate random feedback data."""
    scores = [50, 75, 80, 85, 90, 95]
    tags_sets = [
        ["data_analysis", "enterprise"],
        ["code_generation", "enterprise"],
        ["natural_language_understanding", "enterprise"],
        ["problem_solving", "enterprise"],
        ["communication", "enterprise"],
    ]
    
    capabilities = [
        "tools",
        "tools",
        "tools",
        "tools",
        "tools"
    ]

    capabilities = [
        "data_analysis",
        "code_generation",
        "natural_language_understanding",
        "problem_solving",
        "communication"
    ]
    
    skills = [
        "python",
        "javascript",
        "machine_learning",
        "web_development",
        "cloud_computing"
    ]
    
    return {
        'score': random.choice(scores),
        'tags': random.choice(tags_sets),
        'capability': random.choice(capabilities),
        'skill': random.choice(skills),
        'context': 'enterprise'
    }


def main():
    print("🧪 Testing Agent Feedback Flow with IPFS Pin")
    print_config()
    print("=" * 60)
    
    # SDK Configuration
    sdkConfig_pinata = {
        'chainId': CHAIN_ID,
        'rpcUrl': RPC_URL,
        'ipfs': 'pinata',
        'pinataJwt': PINATA_JWT
        # Subgraph URL auto-defaults from DEFAULT_SUBGRAPH_URLS
    }
    
    # Step 1: Load existing agent
    print("\n📍 Step 1: Load Existing Agent")
    print("-" * 60)
    print(f"Loading agent: {AGENT_ID}")
    
    agentSdk = SDK(**sdkConfig_pinata)  # Read-only for loading
    
    try:
        agent = agentSdk.loadAgent(AGENT_ID)
        print(f"✅ Agent loaded: {agent.name}")
        print(f"   Description: {agent.description[:50]}...")
        print(f"   MCP Endpoint: {agent.mcpEndpoint}")
        print(f"   A2A Endpoint: {agent.a2aEndpoint}")
        print(f"   ENS Endpoint: {agent.ensEndpoint}")
    except Exception as e:
        print(f"❌ Failed to load agent: {e}")
        import traceback
        traceback.print_exc()
        exit(1)
    
    # Step 2: Agent (server) signs feedback auth for client
    print("\n📍 Step 2: Agent (Server) Signs Feedback Auth")
    print("-" * 60)
    
    clientSdk = SDK(signer=CLIENT_PRIVATE_KEY, **sdkConfig_pinata)
    clientAddress = clientSdk.web3_client.account.address
    print(f"Client address: {clientAddress}")
    
    # Agent SDK needs to be initialized with signer for signing feedback auth
    agentSdkWithSigner = SDK(signer=AGENT_PRIVATE_KEY, **sdkConfig_pinata)
    
    # Sign feedback authorization
    print("Signing feedback authorization...")
    feedbackAuth = agentSdkWithSigner.signFeedbackAuth(
        agentId=AGENT_ID,
        clientAddress=clientAddress,
        expiryHours=24
    )
    print(f"✅ Feedback auth signed: {len(feedbackAuth)} bytes")
    
    # Step 3: Client submits feedback
    print("\n📍 Step 3: Client Submits Feedback")
    print("-" * 60)
    
    feedbackEntries = []
    numFeedback = 1
    
    for i in range(numFeedback):
        print(f"\n  Submitting Feedback #{i+1}:")
        feedbackData = generateFeedbackData(i+1)
        
        # Prepare feedback file
        feedbackFile = clientSdk.prepareFeedback(
            agentId=AGENT_ID,
            score=feedbackData['score'],
            tags=feedbackData['tags'],
            capability=feedbackData['capability'],
            skill=feedbackData['skill'],
            context=feedbackData['context']
        )
        
        print(f"  - Score: {feedbackData['score']}/100")
        print(f"  - Tags: {feedbackData['tags']}")
        print(f"  - Capability: {feedbackData['capability']}")
        print(f"  - Skill: {feedbackData['skill']}")
        
        # Submit feedback
        try:
            feedback = clientSdk.giveFeedback(
                agentId=AGENT_ID,
                feedbackFile=feedbackFile,
                feedbackAuth=feedbackAuth
            )
            
            # Extract actual feedback index from the returned Feedback object
            # feedback.id is a tuple: (agentId, clientAddress, feedbackIndex)
            actualFeedbackIndex = feedback.id[2]
            
            feedbackEntries.append({
                'index': actualFeedbackIndex,  # Use actual index from blockchain
                'data': feedbackData,
                'feedback': feedback
            })
            
            print(f"  ✅ Feedback #{actualFeedbackIndex} submitted successfully (entry #{i+1} in this test)")
            if feedback.fileURI:
                print(f"     File URI: {feedback.fileURI}")
            
        except Exception as e:
            print(f"  ❌ Failed to submit feedback #{i+1}: {e}")
            import traceback
            traceback.print_exc()
            exit(1)
        
        time.sleep(2)  # Wait between submissions
    
    # Step 4: Agent (Server) Responds to Feedback
    print("\n📍 Step 4: Agent (Server) Responds to Feedback")
    print("-" * 60)
    
    clientAddress = clientSdk.web3_client.account.address
    
    for i, entry in enumerate(feedbackEntries):
        # Use the actual feedback index that was returned when submitting
        feedbackIndex = entry['index']
        print(f"\n  Responding to Feedback #{feedbackIndex}:")
        
        # Generate response data
        responseData = {
            'text': f"Thank you for your feedback! We appreciate your input.",
            'timestamp': int(time.time()),
            'responder': 'agent'
        }
        
        try:
            # Agent responds to the client's feedback
            updatedFeedback = agentSdkWithSigner.appendResponse(
                agentId=AGENT_ID,
                clientAddress=clientAddress,
                feedbackIndex=feedbackIndex,
                response=responseData
            )
            
            print(f"  ✅ Response submitted to feedback #{feedbackIndex}")
            entry['response'] = responseData
            entry['updatedFeedback'] = updatedFeedback
        except Exception as e:
            print(f"  ❌ Failed to submit response: {e}")
        
        time.sleep(2)  # Wait between responses
    
    # Step 5: Wait for blockchain finalization
    print("\n📍 Step 5: Waiting for Blockchain Finalization")
    print("-" * 60)
    print("⏳ Waiting 15 seconds for blockchain to finalize...")
    time.sleep(15)
    
    # Step 6: Verify feedback data and responses
    print("\n📍 Step 6: Verify Feedback Data Integrity")
    print("-" * 60)
    
    allMatch = True
    
    for i, entry in enumerate(feedbackEntries, 1):
        print(f"\n  Feedback #{i}:")
        data = entry['data']
        feedback = entry['feedback']
        
        # Verify feedback object fields
        checks = [
            ('Score', data['score'], feedback.score),
            ('Tags', data['tags'], feedback.tags),
            ('Capability', data['capability'], feedback.capability),
            ('Skill', data['skill'], feedback.skill),
        ]
        
        for field_name, expected, actual in checks:
            if expected == actual:
                print(f"    ✅ {field_name}: {actual}")
            else:
                print(f"    ❌ {field_name}: expected={expected}, got={actual}")
                allMatch = False
        
        # Verify file URI exists
        if feedback.fileURI:
            print(f"    ✅ File URI: {feedback.fileURI}")
        else:
            print(f"    ⚠️  No file URI (IPFS storage may have failed)")
        
        # Verify server response was added
        if 'response' in entry and entry.get('updatedFeedback'):
            print(f"    ✅ Server Response: Recorded successfully")
    
    # Step 7: Wait for subgraph indexing
    print("\n📍 Step 7: Waiting for Subgraph to Index")
    print("-" * 60)
    print("⏳ Waiting 60 seconds for subgraph to catch up with blockchain events...")
    print("   (Subgraphs can take up to a minute to index new blocks)")
    time.sleep(60)
    
    # Step 8: Test getFeedback (direct access)
    print("\n📍 Step 8: Test getFeedback (Direct Access)")
    print("-" * 60)
    
    for i, entry in enumerate(feedbackEntries):
        # Use the actual feedback index that was returned when submitting
        feedbackIndex = entry['index']
        print(f"\n  Fetching Feedback #{feedbackIndex} using getFeedback():")
        
        try:
            # Use agentSdkWithSigner since agentSdk has no subgraph_client
            retrievedFeedback = agentSdkWithSigner.getFeedback(
                agentId=AGENT_ID,
                clientAddress=clientAddress,
                feedbackIndex=feedbackIndex
            )
            
            print(f"    ✅ Retrieved feedback successfully")
            print(f"    - Score: {retrievedFeedback.score}")
            print(f"    - Tags: {retrievedFeedback.tags}")
            print(f"    - Capability: {retrievedFeedback.capability}")
            print(f"    - Skill: {retrievedFeedback.skill}")
            print(f"    - Is Revoked: {retrievedFeedback.isRevoked}")
            print(f"    - Has Responses: {len(retrievedFeedback.answers)} response(s)")
            if retrievedFeedback.fileURI:
                print(f"    - File URI: {retrievedFeedback.fileURI}")
            
            # Verify retrieved feedback matches original
            expected = entry['data']
            if retrievedFeedback.score == expected['score'] and \
               retrievedFeedback.tags == expected['tags'] and \
               retrievedFeedback.capability == expected['capability'] and \
               retrievedFeedback.skill == expected['skill']:
                print(f"    ✅ Retrieved feedback matches original submission")
            else:
                print(f"    ❌ Retrieved feedback does not match original")
                allMatch = False
                
        except Exception as e:
            print(f"    ❌ Failed to retrieve feedback: {e}")
            allMatch = False
    
    # Step 9: Test searchFeedback (with filters)
    print("\n📍 Step 9: Test searchFeedback (With Filters)")
    print("-" * 60)
    
    # Test 1: Search by capability
    print("\n  Test 1: Search feedback by capability")
    testCapability = feedbackEntries[0]['data']['capability']
    try:
        results = agentSdkWithSigner.searchFeedback(
            agentId=AGENT_ID,
            capabilities=[testCapability],
            first=10,
            skip=0
        )
        print(f"    ✅ Found {len(results)} feedback entry/entries with capability '{testCapability}'")
        if results:
            for fb in results:
                print(f"      - Score: {fb.score}, Tags: {fb.tags}")
    except Exception as e:
        print(f"    ❌ Failed to search feedback by capability: {e}")
        allMatch = False
    
    # Test 2: Search by skill
    print("\n  Test 2: Search feedback by skill")
    testSkill = feedbackEntries[0]['data']['skill']
    try:
        results = agentSdkWithSigner.searchFeedback(
            agentId=AGENT_ID,
            skills=[testSkill],
            first=10,
            skip=0
        )
        print(f"    ✅ Found {len(results)} feedback entry/entries with skill '{testSkill}'")
        if results:
            for fb in results:
                print(f"      - Score: {fb.score}, Tags: {fb.tags}")
    except Exception as e:
        print(f"    ❌ Failed to search feedback by skill: {e}")
        allMatch = False
    
    # Test 3: Search by tags
    print("\n  Test 3: Search feedback by tags")
    testTags = feedbackEntries[0]['data']['tags']
    try:
        results = agentSdkWithSigner.searchFeedback(
            agentId=AGENT_ID,
            tags=testTags,
            first=10,
            skip=0
        )
        print(f"    ✅ Found {len(results)} feedback entry/entries with tags {testTags}")
        if results:
            for fb in results:
                print(f"      - Score: {fb.score}, Capability: {fb.capability}")
    except Exception as e:
        print(f"    ❌ Failed to search feedback by tags: {e}")
        allMatch = False
    
    # Test 4: Search by score range
    print("\n  Test 4: Search feedback by score range (75-95)")
    try:
        results = agentSdkWithSigner.searchFeedback(
            agentId=AGENT_ID,
            minScore=75,
            maxScore=95,
            first=10,
            skip=0
        )
        print(f"    ✅ Found {len(results)} feedback entry/entries with score between 75-95")
        if results:
            scores = sorted([fb.score for fb in results if fb.score])
            print(f"      - Scores found: {scores}")
    except Exception as e:
        print(f"    ❌ Failed to search feedback by score range: {e}")
        allMatch = False
    
    # Final results
    print("\n" + "=" * 60)
    if allMatch:
        print("✅ ALL CHECKS PASSED")
        print("\nSummary:")
        print(f"- Agent ID: {AGENT_ID}")
        print(f"- Agent Name: {agent.name}")
        print(f"- Client address: {clientAddress}")
        print(f"- Feedback entries submitted: {len(feedbackEntries)}")
        print("✅ Feedback flow test complete!")
    else:
        print("❌ SOME CHECKS FAILED")


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"\n❌ Error: {e}")
        import traceback
        traceback.print_exc()
        exit(1)



"""
Main SDK class for Agent0.
"""

from __future__ import annotations

import asyncio
import json
import logging
import time
from typing import Any, Dict, List, Optional, Union, Literal
from datetime import datetime

logger = logging.getLogger(__name__)

from .models import (
    AgentId, ChainId, Address, URI, Timestamp, IdemKey,
    EndpointType, TrustModel, Endpoint, RegistrationFile,
    AgentSummary, Feedback, SearchParams
)
from .web3_client import Web3Client
from .contracts import (
    IDENTITY_REGISTRY_ABI, REPUTATION_REGISTRY_ABI, VALIDATION_REGISTRY_ABI,
    DEFAULT_REGISTRIES, DEFAULT_SUBGRAPH_URLS
)
from .agent import Agent
from .indexer import AgentIndexer
from .ipfs_client import IPFSClient
from .feedback_manager import FeedbackManager
from .subgraph_client import SubgraphClient


class SDK:
    """Main SDK class for Agent0."""

    def __init__(
        self,
        chainId: ChainId,
        rpcUrl: str,
        signer: Optional[Any] = None,  # Optional for read-only operations
        registryOverrides: Optional[Dict[ChainId, Dict[str, Address]]] = None,
        indexingStore: Optional[Any] = None,  # optional (e.g., sqlite/postgres/duckdb)
        embeddings: Optional[Any] = None,  # optional vector backend
        # IPFS configuration
        ipfs: Optional[str] = None,  # "node", "filecoinPin", or "pinata"
        # Direct IPFS node config
        ipfsNodeUrl: Optional[str] = None,
        # Filecoin Pin config
        filecoinPrivateKey: Optional[str] = None,
        # Pinata config
        pinataJwt: Optional[str] = None,
        # Subgraph configuration
        subgraphOverrides: Optional[Dict[ChainId, str]] = None,  # Override subgraph URLs per chain
    ):
        """Initialize the SDK."""
        self.chainId = chainId
        self.rpcUrl = rpcUrl
        self.signer = signer
        
        # Initialize Web3 client (with or without signer for read-only operations)
        if signer:
            if isinstance(signer, str):
                self.web3_client = Web3Client(rpcUrl, private_key=signer)
            else:
                self.web3_client = Web3Client(rpcUrl, account=signer)
        else:
            # Read-only mode - no signer
            self.web3_client = Web3Client(rpcUrl)
        
        # Registry addresses
        self.registry_overrides = registryOverrides or {}
        self._registries = self._resolve_registries()
        
        # Initialize contract instances
        self._identity_registry = None
        self._reputation_registry = None
        self._validation_registry = None
        
        # Resolve subgraph URL (with fallback chain)
        self._subgraph_urls = {}
        if subgraphOverrides:
            self._subgraph_urls.update(subgraphOverrides)
        
        # Get subgraph URL for current chain
        resolved_subgraph_url = None
        
        # Priority 1: Chain-specific override
        if chainId in self._subgraph_urls:
            resolved_subgraph_url = self._subgraph_urls[chainId]
        # Priority 2: Default for chain
        elif chainId in DEFAULT_SUBGRAPH_URLS:
            resolved_subgraph_url = DEFAULT_SUBGRAPH_URLS[chainId]
        else:
            # No subgraph available - subgraph_client will be None
            resolved_subgraph_url = None
        
        # Initialize subgraph client if URL available
        if resolved_subgraph_url:
            self.subgraph_client = SubgraphClient(resolved_subgraph_url)
        else:
            self.subgraph_client = None
        
        # Initialize services
        self.indexer = AgentIndexer(
            web3_client=self.web3_client,
            store=indexingStore,
            embeddings=embeddings,
            subgraph_client=self.subgraph_client,
            subgraph_url_overrides=self._subgraph_urls
        )
        
        # Initialize IPFS client based on configuration
        self.ipfs_client = self._initialize_ipfs_client(
            ipfs, ipfsNodeUrl, filecoinPrivateKey, pinataJwt
        )
        
        # Load registries before passing to FeedbackManager
        identity_registry = self.identity_registry
        reputation_registry = self.reputation_registry
        
        self.feedback_manager = FeedbackManager(
            subgraph_client=self.subgraph_client,
            web3_client=self.web3_client,
            ipfs_client=self.ipfs_client,
            reputation_registry=reputation_registry,
            identity_registry=identity_registry,
            indexer=self.indexer  # Pass indexer for unified search interface
        )

    def _resolve_registries(self) -> Dict[str, Address]:
        """Resolve registry addresses for current chain."""
        # Start with defaults
        registries = DEFAULT_REGISTRIES.get(self.chainId, {}).copy()
        
        # Apply overrides
        if self.chainId in self.registry_overrides:
            registries.update(self.registry_overrides[self.chainId])
        
        return registries

    def _initialize_ipfs_client(
        self, 
        ipfs: Optional[str], 
        ipfsNodeUrl: Optional[str], 
        filecoinPrivateKey: Optional[str], 
        pinataJwt: Optional[str]
    ) -> Optional[IPFSClient]:
        """Initialize IPFS client based on configuration."""
        if not ipfs:
            return None
            
        if ipfs == "node":
            if not ipfsNodeUrl:
                raise ValueError("ipfsNodeUrl is required when ipfs='node'")
            return IPFSClient(url=ipfsNodeUrl, filecoin_pin_enabled=False)
            
        elif ipfs == "filecoinPin":
            if not filecoinPrivateKey:
                raise ValueError("filecoinPrivateKey is required when ipfs='filecoinPin'")
            return IPFSClient(
                url=None, 
                filecoin_pin_enabled=True, 
                filecoin_private_key=filecoinPrivateKey
            )
            
        elif ipfs == "pinata":
            if not pinataJwt:
                raise ValueError("pinataJwt is required when ipfs='pinata'")
            return IPFSClient(
                url=None,
                filecoin_pin_enabled=False,
                pinata_enabled=True,
                pinata_jwt=pinataJwt
            )
            
        else:
            raise ValueError(f"Invalid ipfs value: {ipfs}. Must be 'node', 'filecoinPin', or 'pinata'")

    @property
    def isReadOnly(self) -> bool:
        """Check if SDK is in read-only mode (no signer)."""
        return self.signer is None

    @property
    def identity_registry(self):
        """Get identity registry contract."""
        if self._identity_registry is None:
            address = self._registries.get("IDENTITY")
            if not address:
                raise ValueError(f"No identity registry address for chain {self.chainId}")
            self._identity_registry = self.web3_client.get_contract(
                address, IDENTITY_REGISTRY_ABI
            )
        return self._identity_registry

    @property
    def reputation_registry(self):
        """Get reputation registry contract."""
        if self._reputation_registry is None:
            address = self._registries.get("REPUTATION")
            if not address:
                raise ValueError(f"No reputation registry address for chain {self.chainId}")
            self._reputation_registry = self.web3_client.get_contract(
                address, REPUTATION_REGISTRY_ABI
            )
        return self._reputation_registry

    @property
    def validation_registry(self):
        """Get validation registry contract."""
        if self._validation_registry is None:
            address = self._registries.get("VALIDATION")
            if not address:
                raise ValueError(f"No validation registry address for chain {self.chainId}")
            self._validation_registry = self.web3_client.get_contract(
                address, VALIDATION_REGISTRY_ABI
            )
        return self._validation_registry

    def chain_id(self) -> ChainId:
        """Get current chain ID."""
        return self.chainId

    def registries(self) -> Dict[str, Address]:
        """Get resolved addresses for current chain."""
        return self._registries.copy()

    def get_subgraph_client(self, chain_id: Optional[ChainId] = None) -> Optional[SubgraphClient]:
        """
        Get subgraph client for a specific chain.
        
        Args:
            chain_id: Chain ID (defaults to current chain)
            
        Returns:
            SubgraphClient instance or None if no subgraph available
        """
        target_chain = chain_id if chain_id is not None else self.chainId
        
        # Check if we already have a client for this chain
        if target_chain == self.chainId and self.subgraph_client:
            return self.subgraph_client
        
        # Resolve URL for target chain
        url = None
        if target_chain in self._subgraph_urls:
            url = self._subgraph_urls[target_chain]
        elif target_chain in DEFAULT_SUBGRAPH_URLS:
            url = DEFAULT_SUBGRAPH_URLS[target_chain]
        
        if url:
            return SubgraphClient(url)
        return None

    def set_chain(self, chain_id: ChainId) -> None:
        """Switch chains (advanced)."""
        self.chainId = chain_id
        self._registries = self._resolve_registries()
        # Reset contract instances
        self._identity_registry = None
        self._reputation_registry = None
        self._validation_registry = None

    # Agent lifecycle methods
    def createAgent(
        self,
        name: str,
        description: str,
        image: Optional[URI] = None,
    ) -> Agent:
        """Create a new agent (off-chain object in memory)."""
        registration_file = RegistrationFile(
            name=name,
            description=description,
            image=image,
            updatedAt=int(time.time())
        )
        return Agent(sdk=self, registration_file=registration_file)

    def loadAgent(self, agentId: AgentId) -> Agent:
        """Load an existing agent (hydrates from registration file if registered)."""
        # Convert agentId to string if it's an integer
        agentId = str(agentId)
        
        # Parse agent ID
        if ":" in agentId:
            chain_id, token_id = agentId.split(":", 1)
            if int(chain_id) != self.chainId:
                raise ValueError(f"Agent {agentId} is not on current chain {self.chainId}")
        else:
            token_id = agentId
        
        # Get token URI from contract
        try:
            token_uri = self.web3_client.call_contract(
                self.identity_registry, "tokenURI", int(token_id)
            )
        except Exception as e:
            raise ValueError(f"Failed to load agent {agentId}: {e}")
        
        # Load registration file
        registration_file = self._load_registration_file(token_uri)
        registration_file.agentId = agentId
        registration_file.agentURI = token_uri if token_uri else None
        
        # Store registry address for proper JSON generation
        registry_address = self._registries.get("IDENTITY")
        if registry_address:
            registration_file._registry_address = registry_address
            registration_file._chain_id = self.chainId
        
        # Hydrate on-chain data
        self._hydrate_agent_data(registration_file, int(token_id))
        
        return Agent(sdk=self, registration_file=registration_file)

    def _load_registration_file(self, uri: str) -> RegistrationFile:
        """Load registration file from URI."""
        if uri.startswith("ipfs://"):
            if not self.ipfs_client:
                raise ValueError("IPFS client not configured")
            content = self.ipfs_client.get(uri)
        elif uri.startswith("http"):
            try:
                import requests
                response = requests.get(uri)
                response.raise_for_status()
                content = response.text
            except ImportError:
                raise ImportError("requests not installed. Install with: pip install requests")
        else:
            raise ValueError(f"Unsupported URI scheme: {uri}")
        
        data = json.loads(content)
        return RegistrationFile.from_dict(data)

    def _hydrate_agent_data(self, registration_file: RegistrationFile, token_id: int):
        """Hydrate agent data from on-chain sources."""
        # Get owner
        owner = self.web3_client.call_contract(
            self.identity_registry, "ownerOf", token_id
        )
        registration_file.owners = [owner]
        
        # Get operators (this would require additional contract calls)
        # For now, we'll leave it empty
        registration_file.operators = []
        
        # Hydrate metadata from on-chain (agentWallet, agentName, custom metadata)
        agent_id = token_id
        try:
            # Try to get agentWallet from on-chain metadata
            wallet_bytes = self.web3_client.call_contract(
                self.identity_registry, "getMetadata", agent_id, "agentWallet"
            )
            if wallet_bytes and len(wallet_bytes) > 0:
                wallet_address = "0x" + wallet_bytes.hex()
                registration_file.walletAddress = wallet_address
                # If wallet is read from on-chain, use current chain ID
                # (the chain ID from the registration file might be outdated)
                registration_file.walletChainId = self.chainId
        except Exception as e:
            # No on-chain wallet, will fall back to registration file
            pass
        
        try:
            # Try to get agentName (ENS) from on-chain metadata
            name_bytes = self.web3_client.call_contract(
                self.identity_registry, "getMetadata", agent_id, "agentName"
            )
            if name_bytes and len(name_bytes) > 0:
                ens_name = name_bytes.decode('utf-8')
                # Add ENS endpoint to registration file
                from .models import EndpointType, Endpoint
                # Remove existing ENS endpoints
                registration_file.endpoints = [
                    ep for ep in registration_file.endpoints
                    if ep.type != EndpointType.ENS
                ]
                # Add new ENS endpoint
                ens_endpoint = Endpoint(
                    type=EndpointType.ENS,
                    value=ens_name,
                    meta={"version": "1.0"}
                )
                registration_file.endpoints.append(ens_endpoint)
        except Exception as e:
            # No on-chain ENS name, will fall back to registration file
            pass
        
        # Try to get custom metadata keys from registration file and check on-chain
        # Note: We can't enumerate on-chain metadata keys, so we check each key from the registration file
        # Also check for common custom metadata keys that might exist on-chain
        keys_to_check = list(registration_file.metadata.keys())
        # Also check for known metadata keys that might have been set on-chain
        known_keys = ["testKey", "version", "timestamp", "customField", "anotherField", "numericField"]
        for key in known_keys:
            if key not in keys_to_check:
                keys_to_check.append(key)
        
        for key in keys_to_check:
            try:
                value_bytes = self.web3_client.call_contract(
                    self.identity_registry, "getMetadata", agent_id, key
                )
                if value_bytes and len(value_bytes) > 0:
                    value_str = value_bytes.decode('utf-8')
                    # Try to convert back to original type if possible
                    try:
                        # Try integer
                        value_int = int(value_str)
                        # Check if it's actually stored as integer in metadata or if it was originally a string
                        registration_file.metadata[key] = value_str  # Keep as string for now
                    except ValueError:
                        # Try float
                        try:
                            value_float = float(value_str)
                            registration_file.metadata[key] = value_str  # Keep as string for now
                        except ValueError:
                            registration_file.metadata[key] = value_str
            except Exception as e:
                # Keep registration file value if on-chain not found
                pass

    # Discovery and indexing
    def refreshAgentIndex(self, agentId: AgentId, deep: bool = False) -> AgentSummary:
        """Refresh index for a single agent."""
        return asyncio.run(self.indexer.refresh_agent(agentId, deep=deep))

    def refreshIndex(
        self,
        agentIds: Optional[List[AgentId]] = None,
        concurrency: int = 8,
    ) -> List[AgentSummary]:
        """Refresh index for multiple agents."""
        return asyncio.run(self.indexer.refresh_agents(agentIds, concurrency))

    def getAgent(self, agentId: AgentId) -> AgentSummary:
        """Get agent summary from index."""
        return self.indexer.get_agent(agentId)

    def searchAgents(
        self,
        params: Union[SearchParams, Dict[str, Any], None] = None,
        sort: Union[str, List[str], None] = None,
        page_size: int = 50,
        cursor: Optional[str] = None,
        **kwargs  # Accept search criteria as kwargs for better DX
    ) -> Dict[str, Any]:
        """Search for agents.
        
        Examples:
            # Simple kwargs for better developer experience
            sdk.searchAgents(name="Test")
            sdk.searchAgents(mcpTools=["code_generation"], active=True)
            
            # Explicit SearchParams (for complex queries or IDE autocomplete)
            sdk.searchAgents(SearchParams(name="Test", mcpTools=["code_generation"]))
            
            # With pagination
            sdk.searchAgents(name="Test", page_size=10)
        """
        # If kwargs provided, use them instead of params
        if kwargs and params is None:
            params = SearchParams(**kwargs)
        elif params is None:
            params = SearchParams()
        elif isinstance(params, dict):
            params = SearchParams(**params)
        
        if sort is None:
            sort = ["updatedAt:desc"]
        elif isinstance(sort, str):
            sort = [sort]

        return self.indexer.search_agents(params, sort, page_size, cursor)

    # Feedback methods
    def prepareFeedback(
        self,
        agentId: AgentId,
        score: Optional[int] = None,  # 0-100
        tags: List[str] = None,
        text: Optional[str] = None,
        capability: Optional[str] = None,
        name: Optional[str] = None,
        skill: Optional[str] = None,
        task: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        proofOfPayment: Optional[Dict[str, Any]] = None,
        extra: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Prepare feedback file (local file/object)."""
        return self.feedback_manager.prepareFeedback(
            agentId=agentId,
            score=score,
            tags=tags,
            text=text,
            capability=capability,
            name=name,
            skill=skill,
            task=task,
            context=context,
            proofOfPayment=proofOfPayment,
            extra=extra
        )

    def giveFeedback(
        self,
        agentId: AgentId,
        feedbackFile: Dict[str, Any],
        idem: Optional[IdemKey] = None,
        feedback_auth: Optional[bytes] = None,
    ) -> Feedback:
        """Give feedback (maps 8004 endpoint)."""
        return self.feedback_manager.giveFeedback(
            agentId=agentId,
            feedbackFile=feedbackFile,
            idem=idem,
            feedback_auth=feedback_auth
        )

    def getFeedback(self, feedbackId: str) -> Feedback:
        """Get single feedback by ID string."""
        # Parse feedback ID
        agentId, clientAddress, feedbackIndex = Feedback.from_id_string(feedbackId)
        return self.feedback_manager.getFeedback(agentId, clientAddress, feedbackIndex)

    def searchFeedback(
        self,
        agentId: AgentId,
        reviewers: Optional[List[Address]] = None,
        tags: Optional[List[str]] = None,
        capabilities: Optional[List[str]] = None,
        skills: Optional[List[str]] = None,
        tasks: Optional[List[str]] = None,
        names: Optional[List[str]] = None,
        minScore: Optional[int] = None,
        maxScore: Optional[int] = None,
        include_revoked: bool = False,
        first: int = 100,
        skip: int = 0,
    ) -> List[Feedback]:
        """Search feedback for an agent."""
        return self.feedback_manager.searchFeedback(
            agentId=agentId,
            clientAddresses=reviewers,
            tags=tags,
            capabilities=capabilities,
            skills=skills,
            tasks=tasks,
            names=names,
            minScore=minScore,
            maxScore=maxScore,
            include_revoked=include_revoked,
            first=first,
            skip=skip
        )

    def revokeFeedback(
        self,
        feedbackId: str,
        reason: Optional[str] = None,
        idem: Optional[IdemKey] = None,
    ) -> Dict[str, Any]:
        """Revoke feedback."""
        # Parse feedback ID
        agentId, clientAddress, feedbackIndex = Feedback.from_id_string(feedbackId)
        return self.feedback_manager.revokeFeedback(agentId, feedbackIndex)

    def appendResponse(
        self,
        feedbackId: str,
        response: Dict[str, Any],
        idem: Optional[IdemKey] = None,
    ) -> Feedback:
        """Append a response/follow-up to existing feedback."""
        # Parse feedback ID
        agentId, clientAddress, feedbackIndex = Feedback.from_id_string(feedbackId)
        return self.feedback_manager.appendResponse(agentId, clientAddress, feedbackIndex, response)

    
    def searchAgentsByReputation(
        self,
        agents: Optional[List[AgentId]] = None,
        tags: Optional[List[str]] = None,
        reviewers: Optional[List[Address]] = None,
        capabilities: Optional[List[str]] = None,
        skills: Optional[List[str]] = None,
        tasks: Optional[List[str]] = None,
        names: Optional[List[str]] = None,
        minAverageScore: Optional[int] = None,  # 0-100
        includeRevoked: bool = False,
        page_size: int = 50,
        cursor: Optional[str] = None,
        sort: Optional[List[str]] = None,
        chains: Optional[Union[List[ChainId], Literal["all"]]] = None,
    ) -> Dict[str, Any]:
        """Search agents filtered by reputation criteria."""
        # Handle multi-chain search
        if chains:
            # Expand "all" if needed
            if chains == "all":
                chains = self.indexer._get_all_configured_chains()
            
            # If multiple chains or single chain different from default
            if isinstance(chains, list) and len(chains) > 0:
                if len(chains) > 1 or (len(chains) == 1 and chains[0] != self.chainId):
                    return asyncio.run(
                        self._search_agents_by_reputation_across_chains(
                            agents, tags, reviewers, capabilities, skills, tasks, names,
                            minAverageScore, includeRevoked, page_size, cursor, sort, chains
                        )
                    )
        
        # Single chain search (existing behavior)
        if not self.subgraph_client:
            raise ValueError("Subgraph client required for searchAgentsByReputation")
        
        if sort is None:
            sort = ["createdAt:desc"]
        
        skip = 0
        if cursor:
            try:
                skip = int(cursor)
            except ValueError:
                skip = 0
        
        order_by = "createdAt"
        order_direction = "desc"
        if sort and len(sort) > 0:
            sort_field = sort[0].split(":")
            order_by = sort_field[0] if len(sort_field) >= 1 else order_by
            order_direction = sort_field[1] if len(sort_field) >= 2 else order_direction
        
        try:
            agents_data = self.subgraph_client.search_agents_by_reputation(
                agents=agents,
                tags=tags,
                reviewers=reviewers,
                capabilities=capabilities,
                skills=skills,
                tasks=tasks,
                names=names,
                minAverageScore=minAverageScore,
                includeRevoked=includeRevoked,
                first=page_size,
                skip=skip,
                order_by=order_by,
                order_direction=order_direction
            )
            
            from .models import AgentSummary
            results = []
            for agent_data in agents_data:
                reg_file = agent_data.get('registrationFile') or {}
                if not isinstance(reg_file, dict):
                    reg_file = {}
                
                agent_summary = AgentSummary(
                    chainId=int(agent_data.get('chainId', 0)),
                    agentId=agent_data.get('id'),
                    name=reg_file.get('name', f"Agent {agent_data.get('id')}"),
                    image=reg_file.get('image'),
                    description=reg_file.get('description', ''),
                    owners=[agent_data.get('owner', '')],
                    operators=agent_data.get('operators', []),
                    mcp=reg_file.get('mcpEndpoint') is not None,
                    a2a=reg_file.get('a2aEndpoint') is not None,
                    ens=reg_file.get('ens'),
                    did=reg_file.get('did'),
                    walletAddress=reg_file.get('agentWallet'),
                    supportedTrusts=reg_file.get('supportedTrusts', []),
                    a2aSkills=reg_file.get('a2aSkills', []),
                    mcpTools=reg_file.get('mcpTools', []),
                    mcpPrompts=reg_file.get('mcpPrompts', []),
                    mcpResources=reg_file.get('mcpResources', []),
                    active=reg_file.get('active', True),
                    x402support=reg_file.get('x402support', False),
                    extras={'averageScore': agent_data.get('averageScore')}
                )
                results.append(agent_summary)
            
            next_cursor = str(skip + len(results)) if len(results) == page_size else None
            return {"items": results, "nextCursor": next_cursor}
            
        except Exception as e:
            raise ValueError(f"Failed to search agents by reputation: {e}")
    
    async def _search_agents_by_reputation_across_chains(
        self,
        agents: Optional[List[AgentId]],
        tags: Optional[List[str]],
        reviewers: Optional[List[Address]],
        capabilities: Optional[List[str]],
        skills: Optional[List[str]],
        tasks: Optional[List[str]],
        names: Optional[List[str]],
        minAverageScore: Optional[int],
        includeRevoked: bool,
        page_size: int,
        cursor: Optional[str],
        sort: Optional[List[str]],
        chains: List[ChainId],
    ) -> Dict[str, Any]:
        """
        Search agents by reputation across multiple chains in parallel.
        
        Similar to indexer._search_agents_across_chains() but for reputation-based search.
        """
        import time
        start_time = time.time()
        
        if sort is None:
            sort = ["createdAt:desc"]
        
        order_by = "createdAt"
        order_direction = "desc"
        if sort and len(sort) > 0:
            sort_field = sort[0].split(":")
            order_by = sort_field[0] if len(sort_field) >= 1 else order_by
            order_direction = sort_field[1] if len(sort_field) >= 2 else order_direction
        
        skip = 0
        if cursor:
            try:
                skip = int(cursor)
            except ValueError:
                skip = 0
        
        # Define async function for querying a single chain
        async def query_single_chain(chain_id: int) -> Dict[str, Any]:
            """Query one chain and return its results with metadata."""
            try:
                # Get subgraph client for this chain
                subgraph_client = self.indexer._get_subgraph_client_for_chain(chain_id)
                
                if subgraph_client is None:
                    logger.warning(f"No subgraph client available for chain {chain_id}")
                    return {
                        "chainId": chain_id,
                        "status": "unavailable",
                        "agents": [],
                        "error": f"No subgraph configured for chain {chain_id}"
                    }
                
                # Execute reputation search query
                try:
                    agents_data = subgraph_client.search_agents_by_reputation(
                        agents=agents,
                        tags=tags,
                        reviewers=reviewers,
                        capabilities=capabilities,
                        skills=skills,
                        tasks=tasks,
                        names=names,
                        minAverageScore=minAverageScore,
                        includeRevoked=includeRevoked,
                        first=page_size * 3,  # Fetch extra to allow for filtering/sorting
                        skip=0,  # We'll handle pagination after aggregation
                        order_by=order_by,
                        order_direction=order_direction
                    )
                    
                    logger.info(f"Chain {chain_id}: fetched {len(agents_data)} agents by reputation")
                except Exception as e:
                    logger.error(f"Error in search_agents_by_reputation for chain {chain_id}: {e}", exc_info=True)
                    agents_data = []
                
                return {
                    "chainId": chain_id,
                    "status": "success",
                    "agents": agents_data,
                    "count": len(agents_data),
                }
                
            except Exception as e:
                logger.error(f"Error querying chain {chain_id} for reputation search: {e}", exc_info=True)
                return {
                    "chainId": chain_id,
                    "status": "error",
                    "agents": [],
                    "error": str(e),
                    "count": 0
                }
        
        # Execute queries in parallel
        chain_tasks = [query_single_chain(chain_id) for chain_id in chains]
        chain_results = await asyncio.gather(*chain_tasks)
        
        # Aggregate results from all chains
        all_agents = []
        successful_chains = []
        failed_chains = []
        
        for result in chain_results:
            chain_id = result["chainId"]
            if result["status"] == "success":
                successful_chains.append(chain_id)
                agents_count = len(result.get("agents", []))
                logger.debug(f"Chain {chain_id}: aggregating {agents_count} agents")
                all_agents.extend(result["agents"])
            else:
                failed_chains.append(chain_id)
                logger.warning(f"Chain {chain_id}: status={result.get('status')}, error={result.get('error', 'N/A')}")
        
        logger.debug(f"Total agents aggregated: {len(all_agents)} from {len(successful_chains)} chains")
        
        # Transform to AgentSummary objects
        from .models import AgentSummary
        results = []
        for agent_data in all_agents:
            reg_file = agent_data.get('registrationFile') or {}
            if not isinstance(reg_file, dict):
                reg_file = {}
            
            agent_summary = AgentSummary(
                chainId=int(agent_data.get('chainId', 0)),
                agentId=agent_data.get('id'),
                name=reg_file.get('name', f"Agent {agent_data.get('id')}"),
                image=reg_file.get('image'),
                description=reg_file.get('description', ''),
                owners=[agent_data.get('owner', '')],
                operators=agent_data.get('operators', []),
                mcp=reg_file.get('mcpEndpoint') is not None,
                a2a=reg_file.get('a2aEndpoint') is not None,
                ens=reg_file.get('ens'),
                did=reg_file.get('did'),
                walletAddress=reg_file.get('agentWallet'),
                supportedTrusts=reg_file.get('supportedTrusts', []),
                a2aSkills=reg_file.get('a2aSkills', []),
                mcpTools=reg_file.get('mcpTools', []),
                mcpPrompts=reg_file.get('mcpPrompts', []),
                mcpResources=reg_file.get('mcpResources', []),
                active=reg_file.get('active', True),
                x402support=reg_file.get('x402support', False),
                extras={'averageScore': agent_data.get('averageScore')}
            )
            results.append(agent_summary)
        
        # Sort by averageScore (descending) if available, otherwise by createdAt
        results.sort(
            key=lambda x: (
                x.extras.get('averageScore') if x.extras.get('averageScore') is not None else 0,
                x.chainId,
                x.agentId
            ),
            reverse=True
        )
        
        # Apply pagination
        paginated_results = results[skip:skip + page_size]
        next_cursor = str(skip + len(paginated_results)) if len(paginated_results) == page_size and skip + len(paginated_results) < len(results) else None
        
        elapsed_ms = int((time.time() - start_time) * 1000)
        
        return {
            "items": paginated_results,
            "nextCursor": next_cursor,
            "meta": {
                "chains": chains,
                "successfulChains": successful_chains,
                "failedChains": failed_chains,
                "totalResults": len(results),
                "timing": {"totalMs": elapsed_ms}
            }
        }
    
    # Feedback methods - delegate to feedback_manager
    def signFeedbackAuth(
        self,
        agentId: "AgentId",
        clientAddress: "Address",
        indexLimit: Optional[int] = None,
        expiryHours: int = 24,
    ) -> bytes:
        """Sign feedback authorization for a client."""
        return self.feedback_manager.signFeedbackAuth(
            agentId, clientAddress, indexLimit, expiryHours
        )
    
    def prepareFeedback(
        self,
        agentId: "AgentId",
        score: Optional[int] = None,  # 0-100
        tags: List[str] = None,
        text: Optional[str] = None,
        capability: Optional[str] = None,
        name: Optional[str] = None,
        skill: Optional[str] = None,
        task: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
        proofOfPayment: Optional[Dict[str, Any]] = None,
        extra: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Prepare feedback file (local file/object) according to spec."""
        return self.feedback_manager.prepareFeedback(
            agentId, score, tags, text, capability, name, skill, task, context, proofOfPayment, extra
        )
    
    def giveFeedback(
        self,
        agentId: "AgentId",
        feedbackFile: Dict[str, Any],
        idem: Optional["IdemKey"] = None,
        feedbackAuth: Optional[bytes] = None,
    ) -> "Feedback":
        """Give feedback (maps 8004 endpoint)."""
        return self.feedback_manager.giveFeedback(
            agentId, feedbackFile, idem, feedbackAuth
        )
    
    def getFeedback(
        self,
        agentId: "AgentId",
        clientAddress: "Address",
        feedbackIndex: int,
    ) -> "Feedback":
        """Get feedback (maps 8004 endpoint)."""
        return self.feedback_manager.getFeedback(
            agentId, clientAddress, feedbackIndex
        )
    
    def revokeFeedback(
        self,
        agentId: "AgentId",
        clientAddress: "Address",
        feedbackIndex: int,
    ) -> "Feedback":
        """Revoke feedback."""
        return self.feedback_manager.revokeFeedback(
            agentId, clientAddress, feedbackIndex
        )
    
    def appendResponse(
        self,
        agentId: "AgentId",
        clientAddress: "Address",
        feedbackIndex: int,
        response: Dict[str, Any],
    ) -> "Feedback":
        """Append a response/follow-up to existing feedback."""
        return self.feedback_manager.appendResponse(
            agentId, clientAddress, feedbackIndex, response
        )
    
    def getReputationSummary(
        self,
        agentId: "AgentId",
    ) -> Dict[str, Any]:
        """Get reputation summary for an agent."""
        return self.feedback_manager.getReputationSummary(
            agentId
        )
    
    def transferAgent(
        self,
        agentId: "AgentId",
        newOwnerAddress: str,
    ) -> Dict[str, Any]:
        """Transfer agent ownership to a new address.
        
        Convenience method that loads the agent and calls transfer().
        
        Args:
            agentId: The agent ID to transfer
            newOwnerAddress: Ethereum address of the new owner
            
        Returns:
            Transaction receipt
            
        Raises:
            ValueError: If agent not found or transfer not allowed
        """
        # Load the agent
        agent = self.loadAgent(agentId)
        
        # Call the transfer method
        return agent.transfer(newOwnerAddress)
    
    # Utility methods for owner operations
    def getAgentOwner(self, agentId: AgentId) -> str:
        """Get the current owner of an agent.
        
        Args:
            agentId: The agent ID to check (can be "chainId:tokenId" or just tokenId)
            
        Returns:
            The current owner's Ethereum address
            
        Raises:
            ValueError: If agent ID is invalid or agent doesn't exist
        """
        try:
            # Parse agentId to extract tokenId
            if ":" in str(agentId):
                tokenId = int(str(agentId).split(":")[-1])
            else:
                tokenId = int(agentId)
            
            owner = self.web3_client.call_contract(
                self.identity_registry,
                "ownerOf",
                tokenId
            )
            return owner
        except Exception as e:
            raise ValueError(f"Failed to get owner for agent {agentId}: {e}")
    
    def isAgentOwner(self, agentId: AgentId, address: Optional[str] = None) -> bool:
        """Check if an address is the owner of an agent.
        
        Args:
            agentId: The agent ID to check
            address: Address to check (defaults to SDK's signer address)
            
        Returns:
            True if the address is the owner, False otherwise
            
        Raises:
            ValueError: If agent ID is invalid or agent doesn't exist
        """
        if address is None:
            if not self.signer:
                raise ValueError("No signer available and no address provided")
            address = self.web3_client.account.address
        
        try:
            owner = self.getAgentOwner(agentId)
            return owner.lower() == address.lower()
        except ValueError:
            return False
    
    def canTransferAgent(self, agentId: AgentId, address: Optional[str] = None) -> bool:
        """Check if an address can transfer an agent (i.e., is the owner).
        
        Args:
            agentId: The agent ID to check
            address: Address to check (defaults to SDK's signer address)
            
        Returns:
            True if the address can transfer the agent, False otherwise
        """
        return self.isAgentOwner(agentId, address)



"""
OASF taxonomy validation utilities.
"""

import json
import os
from pathlib import Path
from typing import Optional

# Cache for loaded taxonomy data
_skills_cache: Optional[dict] = None
_domains_cache: Optional[dict] = None


def _get_taxonomy_path(filename: str) -> Path:
    """Get the path to a taxonomy file."""
    # Get the directory where this file is located
    current_dir = Path(__file__).parent
    # Go up one level to agent0_sdk, then into taxonomies
    taxonomy_dir = current_dir.parent / "taxonomies"
    return taxonomy_dir / filename


def _load_skills() -> dict:
    """Load skills taxonomy file with caching."""
    global _skills_cache
    if _skills_cache is None:
        skills_path = _get_taxonomy_path("all_skills.json")
        try:
            with open(skills_path, "r", encoding="utf-8") as f:
                _skills_cache = json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Skills taxonomy file not found: {skills_path}"
            )
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Invalid JSON in skills taxonomy file: {e}"
            )
    return _skills_cache


def _load_domains() -> dict:
    """Load domains taxonomy file with caching."""
    global _domains_cache
    if _domains_cache is None:
        domains_path = _get_taxonomy_path("all_domains.json")
        try:
            with open(domains_path, "r", encoding="utf-8") as f:
                _domains_cache = json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Domains taxonomy file not found: {domains_path}"
            )
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Invalid JSON in domains taxonomy file: {e}"
            )
    return _domains_cache


def validate_skill(slug: str) -> bool:
    """
    Validate if a skill slug exists in the OASF taxonomy.
    
    Args:
        slug: The skill slug to validate (e.g., "natural_language_processing/summarization")
    
    Returns:
        True if the skill exists in the taxonomy, False otherwise
    
    Raises:
        FileNotFoundError: If the taxonomy file cannot be found
        ValueError: If the taxonomy file is invalid JSON
    """
    skills_data = _load_skills()
    skills = skills_data.get("skills", {})
    return slug in skills


def validate_domain(slug: str) -> bool:
    """
    Validate if a domain slug exists in the OASF taxonomy.
    
    Args:
        slug: The domain slug to validate (e.g., "finance_and_business/investment_services")
    
    Returns:
        True if the domain exists in the taxonomy, False otherwise
    
    Raises:
        FileNotFoundError: If the taxonomy file cannot be found
        ValueError: If the taxonomy file is invalid JSON
    """
    domains_data = _load_domains()
    domains = domains_data.get("domains", {})
    return slug in domains




"""
Subgraph client for querying The Graph network.
"""

from __future__ import annotations

import json
import logging
from typing import Any, Dict, List, Optional
import requests

logger = logging.getLogger(__name__)


class SubgraphClient:
    """Client for querying the subgraph GraphQL API."""

    def __init__(self, subgraph_url: str):
        """Initialize subgraph client."""
        self.subgraph_url = subgraph_url

    def query(self, query: str, variables: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Execute a GraphQL query against the subgraph.
        
        Args:
            query: GraphQL query string
            variables: Optional variables for the query
            
        Returns:
            JSON response from the subgraph
        """
        try:
            response = requests.post(
                self.subgraph_url,
                json={
                    'query': query,
                    'variables': variables or {}
                },
                headers={'Content-Type': 'application/json'},
                timeout=10
            )
            response.raise_for_status()
            result = response.json()
            
            # Check for GraphQL errors
            if 'errors' in result:
                error_messages = [err.get('message', 'Unknown error') for err in result['errors']]
                raise ValueError(f"GraphQL errors: {', '.join(error_messages)}")
            
            return result.get('data', {})
        except requests.exceptions.RequestException as e:
            raise ConnectionError(f"Failed to query subgraph: {e}")

    def get_agents(
        self,
        where: Optional[Dict[str, Any]] = None,
        first: int = 100,
        skip: int = 0,
        order_by: str = "createdAt",
        order_direction: str = "desc",
        include_registration_file: bool = True
    ) -> List[Dict[str, Any]]:
        """
        Query agents from the subgraph.
        
        Args:
            where: Filter conditions
            first: Number of results to return
            skip: Number of results to skip
            order_by: Field to order by
            order_direction: Sort direction (asc/desc)
            include_registration_file: Whether to include full registration file data
            
        Returns:
            List of agent records
        """
        # Build WHERE clause
        where_clause = ""
        if where:
            conditions = []
            for key, value in where.items():
                if isinstance(value, bool):
                    conditions.append(f"{key}: {str(value).lower()}")
                elif isinstance(value, str):
                    conditions.append(f'{key}: "{value}"')
                elif isinstance(value, (int, float)):
                    conditions.append(f"{key}: {value}")
                elif isinstance(value, list):
                    conditions.append(f"{key}: {json.dumps(value)}")
            if conditions:
                where_clause = f"where: {{ {', '.join(conditions)} }}"
        
        # Build registration file fragment
        reg_file_fragment = ""
        if include_registration_file:
            reg_file_fragment = """
            registrationFile {
                id
                agentId
                name
                description
                image
                active
                x402support
                supportedTrusts
                mcpEndpoint
                mcpVersion
                a2aEndpoint
                a2aVersion
                ens
                did
                agentWallet
                agentWalletChainId
                mcpTools
                mcpPrompts
                mcpResources
                a2aSkills
                createdAt
            }
            """
        
        query = f"""
        {{
            agents(
                {where_clause}
                first: {first}
                skip: {skip}
                orderBy: {order_by}
                orderDirection: {order_direction}
            ) {{
                id
                chainId
                agentId
                agentURI
                agentURIType
                owner
                operators
                totalFeedback
                createdAt
                updatedAt
                lastActivity
                {reg_file_fragment}
            }}
        }}
        """
        
        result = self.query(query)
        return result.get('agents', [])

    def get_agent_by_id(self, agent_id: str, include_registration_file: bool = True) -> Optional[Dict[str, Any]]:
        """
        Get a specific agent by ID.
        
        Args:
            agent_id: Agent ID in format "chainId:tokenId"
            include_registration_file: Whether to include full registration file data
            
        Returns:
            Agent record or None if not found
        """
        # Build registration file fragment
        reg_file_fragment = ""
        if include_registration_file:
            reg_file_fragment = """
            registrationFile {
                id
                agentId
                name
                description
                image
                active
                x402support
                supportedTrusts
                mcpEndpoint
                mcpVersion
                a2aEndpoint
                a2aVersion
                ens
                did
                agentWallet
                agentWalletChainId
                mcpTools
                mcpPrompts
                mcpResources
                a2aSkills
                createdAt
            }
            """
        
        query = f"""
        {{
            agent(id: "{agent_id}") {{
                id
                chainId
                agentId
                agentURI
                agentURIType
                owner
                operators
                totalFeedback
                createdAt
                updatedAt
                lastActivity
                {reg_file_fragment}
            }}
        }}
        """
        
        result = self.query(query)
        agent = result.get('agent')
        
        if agent is None:
            return None
        
        return agent

    def get_feedback_for_agent(
        self,
        agent_id: str,
        first: int = 100,
        skip: int = 0,
        include_revoked: bool = False
    ) -> List[Dict[str, Any]]:
        """
        Get feedback for a specific agent.
        
        Args:
            agent_id: Agent ID in format "chainId:tokenId"
            first: Number of results to return
            skip: Number of results to skip
            include_revoked: Whether to include revoked feedback
            
        Returns:
            List of feedback records
        """
        query = f"""
        {{
            agent(id: "{agent_id}") {{
                id
                agentId
                feedback(
                    first: {first}
                    skip: {skip}
                    where: {{ isRevoked: {'false' if not include_revoked else 'true'} }}
                    orderBy: createdAt
                    orderDirection: desc
                ) {{
                    id
                    score
                    tag1
                    tag2
                    clientAddress
                    feedbackUri
                    feedbackURIType
                    feedbackHash
                    isRevoked
                    createdAt
                    revokedAt
                    feedbackFile {{
                        id
                        text
                        capability
                        name
                        skill
                        task
                        context
                        proofOfPaymentFromAddress
                        proofOfPaymentToAddress
                        proofOfPaymentChainId
                        proofOfPaymentTxHash
                        tag1
                        tag2
                        createdAt
                    }}
                    responses {{
                        id
                        responder
                        responseUri
                        responseHash
                        createdAt
                    }}
                }}
            }}
        }}
        """
        
        result = self.query(query)
        agent = result.get('agent')
        
        if agent is None:
            return []
        
        return agent.get('feedback', [])

    def get_agent_stats(self, agent_id: str) -> Optional[Dict[str, Any]]:
        """
        Get statistics for a specific agent.
        
        Args:
            agent_id: Agent ID in format "chainId:tokenId"
            
        Returns:
            Agent statistics or None if not found
        """
        query = f"""
        {{
            agentStats(id: "{agent_id}") {{
                agent {{
                    id
                    agentId
                }}
                totalFeedback
                averageScore
                scoreDistribution
                totalValidations
                completedValidations
                averageValidationScore
                lastActivity
                updatedAt
            }}
        }}
        """
        
        result = self.query(query)
        return result.get('agentStats')

    def get_protocol_stats(self, chain_id: int) -> Optional[Dict[str, Any]]:
        """
        Get statistics for a specific protocol/chain.
        
        Args:
            chain_id: Chain ID
            
        Returns:
            Protocol statistics or None if not found
        """
        query = f"""
        {{
            protocol(id: "{chain_id}") {{
                id
                chainId
                name
                identityRegistry
                reputationRegistry
                validationRegistry
                totalAgents
                totalFeedback
                totalValidations
                agents
                tags
                trustModels
                createdAt
                updatedAt
            }}
        }}
        """
        
        result = self.query(query)
        return result.get('protocol')

    def get_global_stats(self) -> Optional[Dict[str, Any]]:
        """
        Get global statistics across all chains.
        
        Returns:
            Global statistics or None if not found
        """
        query = """
        {
            globalStats(id: "stats") {
                totalAgents
                totalFeedback
                totalValidations
                totalProtocols
                agents
                tags
                createdAt
                updatedAt
            }
        }
        """
        
        result = self.query(query)
        return result.get('globalStats')
    
    def get_feedback_by_id(self, feedback_id: str) -> Optional[Dict[str, Any]]:
        """
        Get a specific feedback entry by ID with responses.
        
        Args:
            feedback_id: Feedback ID in format "chainId:agentId:clientAddress:feedbackIndex"
            
        Returns:
            Feedback record with nested feedbackFile and responses, or None if not found
        """
        query = """
        query GetFeedbackById($feedbackId: ID!) {
            feedback(id: $feedbackId) {
                id
                agent { id agentId chainId }
                clientAddress
                score
                tag1
                tag2
                feedbackUri
                feedbackURIType
                feedbackHash
                isRevoked
                createdAt
                revokedAt
                feedbackFile {
                    id
                    feedbackId
                    text
                    capability
                    name
                    skill
                    task
                    context
                    proofOfPaymentFromAddress
                    proofOfPaymentToAddress
                    proofOfPaymentChainId
                    proofOfPaymentTxHash
                    tag1
                    tag2
                    createdAt
                }
                responses {
                    id
                    responder
                    responseUri
                    responseHash
                    createdAt
                }
            }
        }
        """
        variables = {"feedbackId": feedback_id}
        result = self.query(query, variables)
        return result.get('feedback')
    
    def search_feedback(
        self,
        params: Any,  # SearchFeedbackParams
        first: int = 100,
        skip: int = 0,
        order_by: str = "createdAt",
        order_direction: str = "desc",
    ) -> List[Dict[str, Any]]:
        """
        Search for feedback entries with filtering.
        
        Args:
            params: SearchFeedbackParams object with filter criteria
            first: Number of results to return
            skip: Number of results to skip
            order_by: Field to order by
            order_direction: Sort direction (asc/desc)
            
        Returns:
            List of feedback records with nested feedbackFile and responses
        """
        # Build WHERE clause from params
        where_conditions = []
        
        if params.agents is not None and len(params.agents) > 0:
            agent_ids = [f'"{aid}"' for aid in params.agents]
            where_conditions.append(f'agent_in: [{", ".join(agent_ids)}]')
        
        if params.reviewers is not None and len(params.reviewers) > 0:
            reviewers = [f'"{addr}"' for addr in params.reviewers]
            where_conditions.append(f'clientAddress_in: [{", ".join(reviewers)}]')
        
        if not params.includeRevoked:
            where_conditions.append('isRevoked: false')
        
        # Build all non-tag conditions first
        non_tag_conditions = list(where_conditions)
        where_conditions = non_tag_conditions
        
        # Handle tag filtering separately - it needs to be at the top level
        tag_filter_condition = None
        if params.tags is not None and len(params.tags) > 0:
            # Tag search: any of the tags must match in tag1 OR tag2
            # Tags are now stored as human-readable strings in the subgraph
            
            # Build complete condition with all filters for each tag alternative
            # For each tag, create two alternatives: matching tag1 OR matching tag2
            tag_where_items = []
            for tag in params.tags:
                # For tag1 match
                all_conditions_tag1 = non_tag_conditions + [f'tag1: "{tag}"']
                tag_where_items.append(", ".join(all_conditions_tag1))
                # For tag2 match
                all_conditions_tag2 = non_tag_conditions + [f'tag2: "{tag}"']
                tag_where_items.append(", ".join(all_conditions_tag2))
            
            # Join all tag alternatives (each already contains complete filter set)
            tag_filter_condition = ", ".join([f"{{ {item} }}" for item in tag_where_items])
        
        if params.minScore is not None:
            where_conditions.append(f'score_gte: {params.minScore}')
        
        if params.maxScore is not None:
            where_conditions.append(f'score_lte: {params.maxScore}')
        
        # Feedback file filters
        feedback_file_filters = []
        
        if params.capabilities is not None and len(params.capabilities) > 0:
            capabilities = [f'"{cap}"' for cap in params.capabilities]
            feedback_file_filters.append(f'capability_in: [{", ".join(capabilities)}]')
        
        if params.skills is not None and len(params.skills) > 0:
            skills = [f'"{skill}"' for skill in params.skills]
            feedback_file_filters.append(f'skill_in: [{", ".join(skills)}]')
        
        if params.tasks is not None and len(params.tasks) > 0:
            tasks = [f'"{task}"' for task in params.tasks]
            feedback_file_filters.append(f'task_in: [{", ".join(tasks)}]')
        
        if params.names is not None and len(params.names) > 0:
            names = [f'"{name}"' for name in params.names]
            feedback_file_filters.append(f'name_in: [{", ".join(names)}]')
        
        if feedback_file_filters:
            where_conditions.append(f'feedbackFile_: {{ {", ".join(feedback_file_filters)} }}')
        
        # Use tag_filter_condition if tags were provided, otherwise use standard where clause
        if tag_filter_condition:
            # tag_filter_condition already contains properly formatted items: "{ condition1 }, { condition2 }"
            where_clause = f"where: {{ or: [{tag_filter_condition}] }}"
        elif where_conditions:
            where_clause = f"where: {{ {', '.join(where_conditions)} }}"
        else:
            where_clause = ""
        
        query = f"""
        {{
            feedbacks(
                {where_clause}
                first: {first}
                skip: {skip}
                orderBy: {order_by}
                orderDirection: {order_direction}
            ) {{
                id
                agent {{ id agentId chainId }}
                clientAddress
                score
                tag1
                tag2
                feedbackUri
                feedbackURIType
                feedbackHash
                isRevoked
                createdAt
                revokedAt
                feedbackFile {{
                    id
                    feedbackId
                    text
                    capability
                    name
                    skill
                    task
                    context
                    proofOfPaymentFromAddress
                    proofOfPaymentToAddress
                    proofOfPaymentChainId
                    proofOfPaymentTxHash
                    tag1
                    tag2
                    createdAt
                }}
                responses {{
                    id
                    responder
                    responseUri
                    responseHash
                    createdAt
                }}
            }}
        }}
        """
        
        result = self.query(query)
        return result.get('feedbacks', [])
    
    def search_agents_by_reputation(
        self,
        agents: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        reviewers: Optional[List[str]] = None,
        capabilities: Optional[List[str]] = None,
        skills: Optional[List[str]] = None,
        tasks: Optional[List[str]] = None,
        names: Optional[List[str]] = None,
        minAverageScore: Optional[int] = None,  # 0-100
        includeRevoked: bool = False,
        first: int = 100,
        skip: int = 0,
        order_by: str = "createdAt",
        order_direction: str = "desc",
    ) -> List[Dict[str, Any]]:
        """
        Search agents filtered by reputation criteria.
        
        Args:
            agents: List of agent IDs to filter by
            tags: List of tags to filter feedback by
            reviewers: List of reviewer addresses to filter feedback by
            capabilities: List of capabilities to filter feedback by
            skills: List of skills to filter feedback by
            tasks: List of tasks to filter feedback by
            minAverageScore: Minimum average score (0-100) for included agents
            includeRevoked: Whether to include revoked feedback in calculations
            first: Number of results to return
            skip: Number of results to skip
            order_by: Field to order by
            order_direction: Sort direction (asc/desc)
            
        Returns:
            List of agents with averageScore field calculated from filtered feedback
        """
        # Build feedback filter
        feedback_filters = []
        
        if not includeRevoked:
            feedback_filters.append('isRevoked: false')
        
        if tags is not None and len(tags) > 0:
            # Tags are now stored as human-readable strings in the subgraph
            tag_filter = []
            for tag in tags:
                tag_filter.append(f'{{or: [{{tag1: "{tag}"}}, {{tag2: "{tag}"}}]}}')
            feedback_filters.append(f'or: [{", ".join(tag_filter)}]')
        
        if reviewers is not None and len(reviewers) > 0:
            reviewers_list = [f'"{addr}"' for addr in reviewers]
            feedback_filters.append(f'clientAddress_in: [{", ".join(reviewers_list)}]')
        
        # Feedback file filters
        feedback_file_filters = []
        
        if capabilities is not None and len(capabilities) > 0:
            capabilities_list = [f'"{cap}"' for cap in capabilities]
            feedback_file_filters.append(f'capability_in: [{", ".join(capabilities_list)}]')
        
        if skills is not None and len(skills) > 0:
            skills_list = [f'"{skill}"' for skill in skills]
            feedback_file_filters.append(f'skill_in: [{", ".join(skills_list)}]')
        
        if tasks is not None and len(tasks) > 0:
            tasks_list = [f'"{task}"' for task in tasks]
            feedback_file_filters.append(f'task_in: [{", ".join(tasks_list)}]')
        
        if names is not None and len(names) > 0:
            names_list = [f'"{name}"' for name in names]
            feedback_file_filters.append(f'name_in: [{", ".join(names_list)}]')
        
        if feedback_file_filters:
            feedback_filters.append(f'feedbackFile_: {{ {", ".join(feedback_file_filters)} }}')
        
        # If we have feedback filters (tags, capabilities, skills, etc.), we need to first
        # query feedback to get agent IDs, then query those agents
        # Otherwise, query agents directly
        if tags or capabilities or skills or tasks or names or reviewers:
            # First, query feedback to get unique agent IDs that have matching feedback
            feedback_where = f"{{ {', '.join(feedback_filters)} }}" if feedback_filters else "{}"
            
            feedback_query = f"""
            {{
                feedbacks(
                    where: {feedback_where}
                    first: 1000
                    skip: 0
                ) {{
                    agent {{
                        id
                    }}
                }}
            }}
            """
            
            try:
                feedback_result = self.query(feedback_query)
                feedbacks_data = feedback_result.get('feedbacks', [])
                
                # Extract unique agent IDs
                agent_ids_set = set()
                for fb in feedbacks_data:
                    agent = fb.get('agent', {})
                    agent_id = agent.get('id')
                    if agent_id:
                        agent_ids_set.add(agent_id)
                
                if not agent_ids_set:
                    # No agents have matching feedback
                    return []
                
                # Now query only those agents
                agent_ids_list = list(agent_ids_set)
                # Apply any agent filters if specified
                if agents is not None and len(agents) > 0:
                    agent_ids_list = [aid for aid in agent_ids_list if aid in agents]
                    if not agent_ids_list:
                        return []
                
                # Query agents (limit to first N based on pagination)
                agent_ids_str = ', '.join([f'"{aid}"' for aid in agent_ids_list])
                agent_where = f"where: {{ id_in: [{agent_ids_str}] }}"
            except Exception as e:
                logger.warning(f"Failed to query feedback for agent IDs: {e}")
                return []
        else:
            # No feedback filters - query agents directly
            # For reputation search, we want agents that have feedback
            # Filter by totalFeedback > 0 to only get agents with feedback
            agent_filters = ['totalFeedback_gt: 0']  # Only agents with feedback (BigInt comparison)
            if agents is not None and len(agents) > 0:
                agent_ids = [f'"{aid}"' for aid in agents]
                agent_filters.append(f'id_in: [{", ".join(agent_ids)}]')
            
            agent_where = f"where: {{ {', '.join(agent_filters)} }}"
        
        # Build feedback where for agent query (to calculate scores)
        feedback_where_for_agents = f"{{ {', '.join(feedback_filters)} }}" if feedback_filters else "{}"
        
        query = f"""
        {{
            agents(
                {agent_where}
                first: {first}
                skip: {skip}
                orderBy: {order_by}
                orderDirection: {order_direction}
            ) {{
                id
                chainId
                agentId
                agentURI
                agentURIType
                owner
                operators
                createdAt
                updatedAt
                totalFeedback
                lastActivity
                registrationFile {{
                    id
                    name
                    description
                    image
                    active
                    x402support
                    supportedTrusts
                    mcpEndpoint
                    mcpVersion
                    a2aEndpoint
                    a2aVersion
                    ens
                    did
                    agentWallet
                    agentWalletChainId
                    mcpTools
                    mcpPrompts
                    mcpResources
                    a2aSkills
                    createdAt
                }}
                feedback(where: {feedback_where_for_agents}) {{
                    score
                    isRevoked
                    feedbackFile {{
                        capability
                        skill
                        task
                        name
                    }}
                }}
            }}
        }}
        """
        
        try:
            result = self.query(query)
            
            # Check for GraphQL errors
            if 'errors' in result:
                logger.error(f"GraphQL errors in search_agents_by_reputation: {result['errors']}")
                return []
            
            agents_result = result.get('agents', [])
            
            # Calculate average scores
            for agent in agents_result:
                feedbacks = agent.get('feedback', [])
                if feedbacks:
                    scores = [int(fb['score']) for fb in feedbacks if fb.get('score', 0) > 0]
                    if scores:
                        avg_score = sum(scores) / len(scores)
                        agent['averageScore'] = avg_score
                    else:
                        agent['averageScore'] = None
                else:
                    agent['averageScore'] = None
            
            # Filter by minAverageScore
            if minAverageScore is not None:
                agents_result = [
                    agent for agent in agents_result
                    if agent.get('averageScore') is not None and agent['averageScore'] >= minAverageScore
                ]
            
            # For reputation search, filter logic:
            # - If specific agents were requested, return them even if averageScore is None
            #   (the user explicitly asked for these agents, so return them)
            # - If general search (no specific agents), only return agents with reputation data
            if agents is None or len(agents) == 0:
                # General search - only return agents with reputation
                agents_result = [
                    agent for agent in agents_result
                    if agent.get('averageScore') is not None
                ]
            # else: specific agents requested - return all requested agents (even if averageScore is None)
            
            return agents_result
            
        except Exception as e:
            logger.warning(f"Subgraph reputation search failed: {e}")
            return []



#!/usr/bin/env python3
"""
端到端 BNB Greenfield E2E 测试

本脚本简化了 E2E 测试的运行，不依赖 pytest，
直接运行测试来验证 Greenfield 存储的完整工作流程。

使用方法：
python run_greenfield_e2e.py
或：
python run_greenfield_e2e.py all

运行选项：
python run_greenfield_e2e.py small   # 小文本测试
python run_greenfield_e2e.py json    # JSON 数据测试
python run_greenfield_e2e.py binary  # 二进制数据测试
python run_greenfield_e2e.py large   # 大数据性能测试
python run_greenfield_e2e.py all     # 运行所有测试
"""

import asyncio
import json
import os
import sys
import time
import hashlib
import logging
from typing import Any, Dict, Optional
from dotenv import load_dotenv

# Setup logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

# Add project root to Python path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

try:
    from agent0_sdk.core.greenfield_storage import GreenfieldReputationStorage
    from agent0_sdk.core.storage_interfaces import ReputationStorage
except ImportError as e:
    print(f"❌ 导入错误: {e}")
    print("请确保在 agent0-py 目录中运行此脚本，或者已正确安装依赖。")
    sys.exit(1)

# E2E 测试数据
TEST_DATA_EXAMPLES = {
    "small_text": "你好，Greenfield！这是一个 E2E 测试消息。".encode('utf-8'),
    "json_data": json.dumps({
        "agent_id": "test-agent-123",
        "reputation": {
            "score": 95,
            "reviews": [
                {"rating": 5, "comment": "出色的工作"},
                {"rating": 4, "comment": "良好表现"}
            ],
            "created_at": "2024-11-28T14:30:00Z"
        }
    }, ensure_ascii=False).encode('utf-8'),
    "binary_data": bytes([i % 256 for i in range(256)]),  # 包含所有可能的字节值
    "large_data": b"X" * 1024,  # 1KB
    "very_large_data": b"Performance test data. " * 10000,  # ~320KB
}


def check_environment() -> Dict[str, Any]:
    """检查环境配置。"""
    print("🔍 检查环境配置...")

    # 必需字段
    required_fields = [
        "GREENFIELD_BUCKET",
        "GREENFIELD_PRIVATE_KEY",
        "GREENFIELD_SP_HOST"
    ]

    # 可选字段
    optional_fields = [
        "GREENFIELD_TXN_HASH",
        "GREENFIELD_CONTENT_TYPE",
        "GREENFIELD_TIMEOUT"
    ]

    config = {}
    missing_fields = []

    for field in required_fields:
        value = os.getenv(field)
        if value:
            config[field] = value
        else:
            missing_fields.append(field)

    for field in optional_fields:
        value = os.getenv(field)
        if value:
            config[field] = value

    if missing_fields:
        print("❌ 缺少必需的环境变量：")
        for field in missing_fields:
            print(f"   - {field}")
        print(f"\n📖 请参考 .env 文件配置环境变量。")
        return None

    print(f"✅ 环境配置检查通过")
    print(f"   - Bucket: {config['GREENFIELD_BUCKET']}")
    print(f"   - SP Host: {config['GREENFIELD_SP_HOST']}")
    print(f"   - Has Txn Hash: {'GREENFIELD_TXN_HASH' in config}")
    return config


def create_greenfield_storage(config: Dict[str, Any]) -> GreenfieldReputationStorage:
    """创建 Greenfield 存储实例。"""
    try:
        storage = GreenfieldReputationStorage(
            sp_host=config["GREENFIELD_SP_HOST"],
            bucket=config["GREENFIELD_BUCKET"],
            private_key=config["GREENFIELD_PRIVATE_KEY"],
            txn_hash=config.get("GREENFIELD_TXN_HASH"),
            content_type=config.get("GREENFIELD_CONTENT_TYPE", "application/octet-stream"),
            timeout=int(config.get("GREENFIELD_TIMEOUT", 30))
        )
        print(f"✅ Greenfield 存储创建成功")
        return storage
    except Exception as e:
        print(f"❌ 创建 Greenfield 存储失败: {e}")
        raise


async def test_upload_download(storage: GreenfieldReputationStorage, test_name: str, data: bytes, key: Optional[str] = None) -> bool:
    """测试上传和下载的完整流程。"""
    print(f"\n🚀 开始测试: {test_name}")
    print(f"   数据大小: {len(data)} 字节")

    try:
        # 上传数据
        start_time = time.time()
        object_key = storage.put(key, data)
        upload_time = time.time() - start_time

        print(f"✅ 上传成功:")
        print(f"   对象键: {object_key}")
        print(f"   上传时间: {upload_time:.2f} 秒")
        print(f"   上传速度: {len(data) / upload_time:.2f} 字节/秒")

        # 下载数据
        start_time = time.time()
        downloaded_data = storage.get(object_key)
        download_time = time.time() - start_time

        print(f"✅ 下载成功:")
        print(f"   下载大小: {len(downloaded_data)} 字节")
        print(f"   下载时间: {download_time:.2f} 秒")
        print(f"   下载速度: {len(downloaded_data) / download_time:.2f} 字节/秒")

        # 验证数据完整性
        if downloaded_data == data:
            print(f"✅ 数据完整性验证通过")

            # 计算哈希值
            original_hash = hashlib.sha256(data).hexdigest()
            downloaded_hash = hashlib.sha256(downloaded_data).hexdigest()
            print(f"   原始数据 SHA256: {original_hash[:16]}...")
            print(f"   下载数据 SHA256: {downloaded_hash[:16]}...")

            return True
        else:
            print(f"❌ 数据完整性验证失败")
            print(f"   原始数据长度: {len(data)}")
            print(f"   下载数据长度: {len(downloaded_data)}")
            return False

    except Exception as e:
        print(f"❌ 测试失败: {e}")
        return False


async def test_error_handling(storage: GreenfieldReputationStorage) -> bool:
    """测试错误处理。"""
    print(f"\n🧪 开始错误处理测试")

    try:
        # 测试获取不存在的对象
        print("   测试获取不存在的对象...")
        try:
            storage.get("non-existent-object-key-12345")
            print("❌ 应该抛出异常但没有")
            return False
        except RuntimeError as e:
            print(f"✅ 正确抛出异常: {type(e).__name__}")

        # 测试上传时缺少 txn_hash
        print("   测试上传时缺少 txn_hash...")
        # 创建一个没有默认 txn_hash 的存储实例
        try:
            storage_no_txn = GreenfieldReputationStorage(
                sp_host=os.getenv("GREENFIELD_SP_HOST"),
                bucket=os.getenv("GREENFIELD_BUCKET"),
                private_key=os.getenv("GREENFIELD_PRIVATE_KEY"),
                txn_hash=None  # 没有默认 txn_hash
            )
            storage_no_txn.put("test-key", b"test data")
            print("❌ 应该抛出异常但没有")
            return False
        except ValueError as e:
            print(f"✅ 正确抛出异常: {type(e).__name__}")
            print(f"   错误信息: {str(e)[:100]}...")

        print(f"✅ 错误处理测试通过")
        return True

    except Exception as e:
        print(f"❌ 错误处理测试失败: {e}")
        return False


async def test_performance(storage: GreenfieldReputationStorage) -> bool:
    """测试性能。"""
    print(f"\n⚡ 开始性能测试")

    try:
        # 测试不同大小的数据
        test_sizes = [
            (b"Small test", "小数据"),
            (b"X" * 1024, "1KB 数据"),
            (b"Y" * 10240, "10KB 数据"),
            (b"Z" * 102400, "100KB 数据"),
        ]

        performance_results = []

        for data, description in test_sizes:
            print(f"   测试 {description} ({len(data)} 字节)...")

            # 上传性能测试
            start_time = time.time()
            object_key = storage.put(None, data)
            upload_time = time.time() - start_time

            # 下载性能测试
            start_time = time.time()
            downloaded_data = storage.get(object_key)
            download_time = time.time() - start_time

            # 计算速度
            upload_speed = len(data) / upload_time
            download_speed = len(downloaded_data) / download_time

            performance_results.append({
                "description": description,
                "size": len(data),
                "upload_time": upload_time,
                "download_time": download_time,
                "upload_speed": upload_speed,
                "download_speed": download_speed
            })

            print(f"     上传: {upload_time:.2f}s ({upload_speed:.2f} B/s)")
            print(f"     下载: {download_time:.2f}s ({download_speed:.2f} B/s)")

        # 显示性能总结
        print(f"\n📊 性能总结:")
        print(f"{'描述':<15} {'大小':<10} {'上传速度':<15} {'下载速度':<15}")
        print("-" * 60)
        for result in performance_results:
            print(f"{result['description']:<15} {result['size']:<10} "
                  f"{result['upload_speed']:<15.2f} {result['download_speed']:<15.2f}")

        print(f"✅ 性能测试完成")
        return True

    except Exception as e:
        print(f"❌ 性能测试失败: {e}")
        return False


async def run_all_tests(storage: GreenfieldReputationStorage) -> Dict[str, bool]:
    """运行所有测试。"""
    print(f"\n🧪 开始运行完整的 E2E 测试套件")

    results = {}

    # 基础数据测试
    for test_name, data in TEST_DATA_EXAMPLES.items():
        if test_name == "very_large_data":
            continue  # 单独测试大数据
        results[test_name] = await test_upload_download(storage, test_name, data)

    # 自动生成键测试
    print(f"\n🔑 测试自动生成对象键...")
    results["auto_key"] = await test_upload_download(storage, "自动键生成", b"Auto-generated key test", None)

    # 指定键测试
    print(f"\n🏷️ 测试指定对象键...")
    results["manual_key"] = await test_upload_download(storage, "手动键", b"Manual key test", "test-manual-key-123")

    # 错误处理测试
    results["error_handling"] = await test_error_handling(storage)

    # 性能测试
    results["performance"] = await test_performance(storage)

    # 大数据测试（如果环境允许）
    print(f"\n📦 测试大数据 (~320KB)...")
    results["very_large_data"] = await test_upload_download(storage, "大数据测试", TEST_DATA_EXAMPLES["very_large_data"])

    return results


def print_summary(results: Dict[str, bool]) -> None:
    """打印测试结果总结。"""
    print(f"\n📋 测试结果总结")
    print("=" * 50)

    passed = sum(1 for result in results.values() if result)
    total = len(results)

    for test_name, result in results.items():
        status = "✅ 通过" if result else "❌ 失败"
        print(f"{test_name:<25} {status}")

    print("-" * 50)
    print(f"总计: {passed}/{total} 测试通过")

    if passed == total:
        print(f"🎉 所有测试通过！Greenfield 存储工作正常。")
    else:
        print(f"⚠️ 有 {total - passed} 个测试失败，请检查配置和网络连接。")


async def main():
    """主函数。"""
    print("🌟 BNB Greenfield E2E 测试")
    print("=" * 50)

    # 加载环境变量
    load_dotenv()

    # 解析命令行参数
    test_type = sys.argv[1] if len(sys.argv) > 1 else "all"

    # 检查环境配置
    config = check_environment()
    if not config:
        sys.exit(1)

    # 创建存储实例
    try:
        storage = create_greenfield_storage(config)
    except Exception:
        sys.exit(1)

    # 运行测试
    try:
        if test_type == "all":
            results = await run_all_tests(storage)
        elif test_type == "small":
            results = {"small_text": await test_upload_download(storage, "小文本测试", TEST_DATA_EXAMPLES["small_text"])}
        elif test_type == "json":
            results = {"json_data": await test_upload_download(storage, "JSON数据测试", TEST_DATA_EXAMPLES["json_data"])}
        elif test_type == "binary":
            results = {"binary_data": await test_upload_download(storage, "二进制数据测试", TEST_DATA_EXAMPLES["binary_data"])}
        elif test_type == "large":
            results = {"large_data": await test_upload_download(storage, "大数据测试", TEST_DATA_EXAMPLES["large_data"])}
        else:
            print(f"❌ 未知的测试类型: {test_type}")
            print("可用选项: small, json, binary, large, all")
            sys.exit(1)

        # 打印总结
        print_summary(results)

        # 返回适当的退出码
        sys.exit(0 if all(results.values()) else 1)

    except KeyboardInterrupt:
        print(f"\n⚠️ 测试被用户中断")
        sys.exit(1)
    except Exception as e:
        print(f"\n❌ 测试运行失败: {e}")
        logger.exception("Test execution failed")
        sys.exit(1)


if __name__ == "__main__":
    asyncio.run(main())


"""
Tests for OASF skills and domains management.
"""

import pytest
from unittest.mock import patch, MagicMock

from agent0_sdk.core.sdk import SDK
from agent0_sdk.core.models import EndpointType
from agent0_sdk.core.oasf_validator import validate_skill, validate_domain


class TestOASFManagement:
    """Test OASF skills and domains management."""

    def test_add_skill_without_validation(self):
        """Test adding a skill without validation."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            # Create a mock instance with chain_id attribute
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add skill without validation
            agent.addSkill("custom_skill/test_skill", validate_oasf=False)
            
            # Verify OASF endpoint was created
            oasf_endpoints = [ep for ep in agent.endpoints if ep.type == EndpointType.OASF]
            assert len(oasf_endpoints) == 1
            
            oasf_endpoint = oasf_endpoints[0]
            assert oasf_endpoint.value == "https://github.com/agntcy/oasf/"
            assert oasf_endpoint.meta["version"] == "v0.8.0"
            assert "custom_skill/test_skill" in oasf_endpoint.meta["skills"]

    def test_add_skill_with_validation_valid(self):
        """Test adding a valid skill with validation."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add valid skill with validation
            agent.addSkill("advanced_reasoning_planning/strategic_planning", validate_oasf=True)
            
            # Verify skill was added
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert "advanced_reasoning_planning/strategic_planning" in oasf_endpoint.meta["skills"]

    def test_add_skill_with_validation_invalid(self):
        """Test adding an invalid skill with validation raises ValueError."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Try to add invalid skill with validation
            with pytest.raises(ValueError, match="Invalid OASF skill slug"):
                agent.addSkill("invalid_skill/does_not_exist", validate_oasf=True)

    def test_add_skill_duplicate(self):
        """Test adding duplicate skill doesn't create duplicates."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add same skill twice
            agent.addSkill("test_skill/slug", validate_oasf=False)
            agent.addSkill("test_skill/slug", validate_oasf=False)
            
            # Verify only one instance exists
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert oasf_endpoint.meta["skills"].count("test_skill/slug") == 1

    def test_remove_skill_existing(self):
        """Test removing an existing skill."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add skill
            agent.addSkill("test_skill/slug1", validate_oasf=False)
            agent.addSkill("test_skill/slug2", validate_oasf=False)
            
            # Remove one skill
            agent.removeSkill("test_skill/slug1")
            
            # Verify skill was removed
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert "test_skill/slug1" not in oasf_endpoint.meta["skills"]
            assert "test_skill/slug2" in oasf_endpoint.meta["skills"]

    def test_remove_skill_non_existent(self):
        """Test removing a non-existent skill succeeds silently."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Remove skill when OASF endpoint doesn't exist (should succeed silently)
            agent.removeSkill("non_existent_skill")
            
            # Should not raise an error

    def test_remove_skill_when_endpoint_missing(self):
        """Test removing skill when OASF endpoint doesn't exist."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Remove skill when no OASF endpoint exists
            agent.removeSkill("some_skill")
            
            # Should succeed silently (idempotent)

    def test_add_domain_without_validation(self):
        """Test adding a domain without validation."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add domain without validation
            agent.addDomain("custom_domain/test_domain", validate_oasf=False)
            
            # Verify OASF endpoint was created
            oasf_endpoints = [ep for ep in agent.endpoints if ep.type == EndpointType.OASF]
            assert len(oasf_endpoints) == 1
            
            oasf_endpoint = oasf_endpoints[0]
            assert "custom_domain/test_domain" in oasf_endpoint.meta["domains"]

    def test_add_domain_with_validation_valid(self):
        """Test adding a valid domain with validation."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add valid domain with validation
            agent.addDomain("finance_and_business/investment_services", validate_oasf=True)
            
            # Verify domain was added
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert "finance_and_business/investment_services" in oasf_endpoint.meta["domains"]

    def test_add_domain_with_validation_invalid(self):
        """Test adding an invalid domain with validation raises ValueError."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Try to add invalid domain with validation
            with pytest.raises(ValueError, match="Invalid OASF domain slug"):
                agent.addDomain("invalid_domain/does_not_exist", validate_oasf=True)

    def test_add_domain_duplicate(self):
        """Test adding duplicate domain doesn't create duplicates."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add same domain twice
            agent.addDomain("test_domain/slug", validate_oasf=False)
            agent.addDomain("test_domain/slug", validate_oasf=False)
            
            # Verify only one instance exists
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert oasf_endpoint.meta["domains"].count("test_domain/slug") == 1

    def test_remove_domain_existing(self):
        """Test removing an existing domain."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add domains
            agent.addDomain("test_domain/slug1", validate_oasf=False)
            agent.addDomain("test_domain/slug2", validate_oasf=False)
            
            # Remove one domain
            agent.removeDomain("test_domain/slug1")
            
            # Verify domain was removed
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert "test_domain/slug1" not in oasf_endpoint.meta["domains"]
            assert "test_domain/slug2" in oasf_endpoint.meta["domains"]

    def test_remove_domain_non_existent(self):
        """Test removing a non-existent domain succeeds silently."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Remove domain when OASF endpoint doesn't exist (should succeed silently)
            agent.removeDomain("non_existent_domain")
            
            # Should not raise an error

    def test_method_chaining(self):
        """Test method chaining."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Chain multiple operations
            agent.addSkill("skill1", validate_oasf=False)\
                 .addDomain("domain1", validate_oasf=False)\
                 .addSkill("skill2", validate_oasf=False)\
                 .removeSkill("skill1")
            
            # Verify results
            oasf_endpoint = next(ep for ep in agent.endpoints if ep.type == EndpointType.OASF)
            assert "skill1" not in oasf_endpoint.meta["skills"]
            assert "skill2" in oasf_endpoint.meta["skills"]
            assert "domain1" in oasf_endpoint.meta["domains"]

    def test_serialization_deserialization(self):
        """Test that OASF data persists through serialization/deserialization."""
        with patch('agent0_sdk.core.sdk.Web3Client') as mock_web3_class:
            mock_web3_instance = MagicMock()
            mock_web3_instance.chain_id = 11155111
            mock_web3_class.return_value = mock_web3_instance
            
            sdk = SDK(
                chainId=11155111,
                signer="0x1234567890abcdef1234567890abcdef12345678",
                rpcUrl="https://eth-sepolia.g.alchemy.com/v2/test"
            )
            
            agent = sdk.createAgent("Test Agent", "A test agent")
            
            # Add skills and domains
            agent.addSkill("test_skill/slug1", validate_oasf=False)
            agent.addSkill("test_skill/slug2", validate_oasf=False)
            agent.addDomain("test_domain/slug1", validate_oasf=False)
            
            # Serialize to dict
            reg_file_dict = agent.registration_file.to_dict()
            
            # Find OASF endpoint in serialized data
            oasf_endpoint_dict = next(
                ep for ep in reg_file_dict["endpoints"] 
                if ep["name"] == "OASF"
            )
            
            # Verify data is present
            assert "test_skill/slug1" in oasf_endpoint_dict["skills"]
            assert "test_skill/slug2" in oasf_endpoint_dict["skills"]
            assert "test_domain/slug1" in oasf_endpoint_dict["domains"]
            
            # Deserialize back
            from agent0_sdk.core.models import RegistrationFile
            new_reg_file = RegistrationFile.from_dict(reg_file_dict)
            
            # Verify OASF endpoint was restored
            oasf_endpoints = [ep for ep in new_reg_file.endpoints if ep.type == EndpointType.OASF]
            assert len(oasf_endpoints) == 1
            
            oasf_endpoint = oasf_endpoints[0]
            assert "test_skill/slug1" in oasf_endpoint.meta["skills"]
            assert "test_skill/slug2" in oasf_endpoint.meta["skills"]
            assert "test_domain/slug1" in oasf_endpoint.meta["domains"]


class TestOASFValidator:
    """Test OASF validator utilities."""

    def test_validate_skill_valid(self):
        """Test validating a valid skill."""
        assert validate_skill("advanced_reasoning_planning/strategic_planning") is True

    def test_validate_skill_invalid(self):
        """Test validating an invalid skill."""
        assert validate_skill("invalid_skill/does_not_exist") is False

    def test_validate_domain_valid(self):
        """Test validating a valid domain."""
        assert validate_domain("finance_and_business/investment_services") is True

    def test_validate_domain_invalid(self):
        """Test validating an invalid domain."""
        assert validate_domain("invalid_domain/does_not_exist") is False




"""
End-to-End Greenfield E2E Test

This test demonstrates the complete Greenfield workflow:
1. CreateObject on-chain (automatically via helper)
2. PutObject upload to storage provider
3. GetObject download verification
4. Cross-key retrieval to verify public access

To run this test:
1. Set environment variables (see GREENFIELD_E2E_GUIDE.md or GREENFIELD_E2E_使用指南.md)
2. Run: pytest tests/test_greenfield_e2e.py -v -m integration -s

Environment variables required:
- GREENFIELD_RPC_URL (default: https://gnfd-testnet-fullnode-tendermint-us.bnbchain.org)
- GREENFIELD_SP_HOST (default: gnfd-testnet-sp1.bnbchain.org)
- GREENFIELD_BUCKET
- GREENFIELD_PRIVATE_KEY
- Optional: GREENFIELD_CHAIN_ID (default: 5600 for testnet)

For Chinese documentation, see GREENFIELD_E2E_使用指南.md
For English documentation, see GREENFIELD_E2E_GUIDE.md
"""


"""
Storage interfaces for reputation data.

This module defines abstract interfaces for reputation storage backends,
allowing multiple implementations (IPFS, Greenfield, etc.) to be swapped transparently.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, Optional


class ReputationStorage(ABC):
    """Abstract interface for reputation data storage.

    This interface defines the contract for storing and retrieving reputation data,
    allowing multiple backend implementations (IPFS, BNB Greenfield, etc.) to be used
    interchangeably without affecting upper-layer business logic.
    """

    @abstractmethod
    def put(self, key: str, data: bytes, txn_hash: Optional[str] = None) -> str:
        """Store data and return a unique identifier.

        Args:
            key: Unique key for the data (can be empty string for auto-generation)
            data: Binary data to store
            txn_hash: Optional transaction hash for Greenfield CreateObject operation.
                     Required for Greenfield, ignored for IPFS.
                     If not provided, uses the default from constructor (if available).

        Returns:
            Unique identifier (CID for IPFS, object key for Greenfield)
        """
        pass

    @abstractmethod
    def get(self, key: str) -> bytes:
        """Retrieve data by key.

        Args:
            key: Unique identifier returned by put()

        Returns:
            Binary data

        Raises:
            RuntimeError: If data cannot be retrieved
        """
        pass



"""
Agent0 SDK - Python SDK for agent portability, discovery and trust based on ERC-8004.
"""

from .core.models import (
    AgentId,
    ChainId,
    Address,
    URI,
    CID,
    Timestamp,
    IdemKey,
    EndpointType,
    TrustModel,
    Endpoint,
    RegistrationFile,
    AgentSummary,
    Feedback,
    SearchParams,
    SearchFeedbackParams,
)

# Try to import SDK and Agent (may fail if web3 is not installed)
try:
    from .core.sdk import SDK
    from .core.agent import Agent
    _sdk_available = True
except ImportError:
    SDK = None
    Agent = None
    _sdk_available = False

__version__ = "0.31"
__all__ = [
    "SDK",
    "Agent",
    "AgentId",
    "ChainId", 
    "Address",
    "URI",
    "CID",
    "Timestamp",
    "IdemKey",
    "EndpointType",
    "TrustModel",
    "Endpoint",
    "RegistrationFile",
    "AgentSummary",
    "Feedback",
    "SearchParams",
    "SearchFeedbackParams",
]



"""
Core data models for the Agent0 SDK.
"""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union, Literal
from datetime import datetime


# Type aliases
AgentId = str  # "chainId:tokenId" (e.g., "8453:1234") or just tokenId when chain is implicit
ChainId = int
Address = str  # 0x-hex
URI = str  # https://... or ipfs://...
CID = str  # IPFS CID (if used)
Timestamp = int  # unix seconds
IdemKey = str  # idempotency key for write ops


class EndpointType(Enum):
    """Types of endpoints that agents can advertise."""
    MCP = "MCP"
    A2A = "A2A"
    ENS = "ENS"
    DID = "DID"
    WALLET = "wallet"
    OASF = "OASF"


class TrustModel(Enum):
    """Trust models supported by the SDK."""
    REPUTATION = "reputation"
    CRYPTO_ECONOMIC = "crypto-economic"
    TEE_ATTESTATION = "tee-attestation"


@dataclass
class Endpoint:
    """Represents an agent endpoint."""
    type: EndpointType
    value: str  # endpoint value (URL, name, DID, ENS)
    meta: Dict[str, Any] = field(default_factory=dict)  # optional metadata


@dataclass
class RegistrationFile:
    """Agent registration file structure."""
    agentId: Optional[AgentId] = None  # None until minted
    agentURI: Optional[URI] = None  # where this file is (or will be) published
    name: str = ""
    description: str = ""
    image: Optional[URI] = None
    walletAddress: Optional[Address] = None
    walletChainId: Optional[int] = None  # Chain ID for the wallet address
    endpoints: List[Endpoint] = field(default_factory=list)
    trustModels: List[Union[TrustModel, str]] = field(default_factory=list)
    owners: List[Address] = field(default_factory=list)  # from chain (read-only, hydrated)
    operators: List[Address] = field(default_factory=list)  # from chain (read-only, hydrated)
    active: bool = False  # SDK extension flag
    x402support: bool = False  # Binary flag for x402 payment support
    metadata: Dict[str, Any] = field(default_factory=dict)  # arbitrary, SDK-managed
    updatedAt: Timestamp = field(default_factory=lambda: int(datetime.now().timestamp()))

    def __str__(self) -> str:
        """String representation as JSON."""
        # Use stored registry info if available
        chain_id = getattr(self, '_chain_id', None)
        registry_address = getattr(self, '_registry_address', None)
        return json.dumps(self.to_dict(chain_id, registry_address), indent=2, default=str)
    
    def __repr__(self) -> str:
        """Developer representation."""
        return f"RegistrationFile(agentId={self.agentId}, agentURI={self.agentURI}, name={self.name})"
    
    def to_dict(self, chain_id: Optional[int] = None, identity_registry_address: Optional[str] = None) -> Dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        # Build endpoints array
        endpoints = []
        for endpoint in self.endpoints:
            endpoint_dict = {
                "name": endpoint.type.value,
                "endpoint": endpoint.value,
                **endpoint.meta
            }
            endpoints.append(endpoint_dict)
        
        # Add walletAddress as an endpoint if present
        if self.walletAddress:
            # Use stored walletChainId if available, otherwise extract from agentId
            chain_id_for_wallet = self.walletChainId
            if chain_id_for_wallet is None:
                # Extract chain ID from agentId if available, otherwise use 1 as default
                chain_id_for_wallet = 1  # Default to mainnet
                if self.agentId and ":" in self.agentId:
                    try:
                        chain_id_for_wallet = int(self.agentId.split(":")[1])
                    except (ValueError, IndexError):
                        chain_id_for_wallet = 1
            
            endpoints.append({
                "name": "agentWallet",
                "endpoint": f"eip155:{chain_id_for_wallet}:{self.walletAddress}"
            })
        
        # Build registrations array
        registrations = []
        if self.agentId:
            agent_id_int = int(self.agentId.split(":")[-1]) if ":" in self.agentId else int(self.agentId)
            agent_registry = f"eip155:{chain_id}:{identity_registry_address}" if chain_id and identity_registry_address else f"eip155:1:{{identityRegistry}}"
            registrations.append({
                "agentId": agent_id_int,
                "agentRegistry": agent_registry
            })
        
        return {
            "type": "https://eips.ethereum.org/EIPS/eip-8004#registration-v1",
            "name": self.name,
            "description": self.description,
            "image": self.image,
            "endpoints": endpoints,
            "registrations": registrations,
            "supportedTrust": [tm.value if isinstance(tm, TrustModel) else tm for tm in self.trustModels],
            "active": self.active,
            "x402support": self.x402support,
            "updatedAt": self.updatedAt,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> RegistrationFile:
        """Create from dictionary."""
        endpoints = []
        for ep_data in data.get("endpoints", []):
            name = ep_data["name"]
            # Special handling for agentWallet - it's not a standard endpoint type
            if name == "agentWallet":
                # Skip agentWallet endpoints as they're handled separately via walletAddress field
                continue
            
            ep_type = EndpointType(name)
            ep_value = ep_data["endpoint"]
            ep_meta = {k: v for k, v in ep_data.items() if k not in ["name", "endpoint"]}
            endpoints.append(Endpoint(type=ep_type, value=ep_value, meta=ep_meta))

        trust_models = []
        for tm in data.get("supportedTrust", []):
            try:
                trust_models.append(TrustModel(tm))
            except ValueError:
                trust_models.append(tm)  # custom string

        return cls(
            agentId=data.get("agentId"),
            agentURI=data.get("agentURI"),
            name=data.get("name", ""),
            description=data.get("description", ""),
            image=data.get("image"),
            walletAddress=data.get("walletAddress"),
            walletChainId=data.get("walletChainId"),
            endpoints=endpoints,
            trustModels=trust_models,
            active=data.get("active", False),
            x402support=data.get("x402support", False),
            metadata=data.get("metadata", {}),
            updatedAt=data.get("updatedAt", int(datetime.now().timestamp())),
        )


@dataclass
class AgentSummary:
    """Summary information for agent discovery and search."""
    chainId: ChainId
    agentId: AgentId
    name: str
    image: Optional[URI]
    description: str
    owners: List[Address]
    operators: List[Address]
    mcp: bool
    a2a: bool
    ens: Optional[str]
    did: Optional[str]
    walletAddress: Optional[Address]
    supportedTrusts: List[str]  # normalized string keys
    a2aSkills: List[str]
    mcpTools: List[str]
    mcpPrompts: List[str]
    mcpResources: List[str]
    active: bool
    x402support: bool = False
    extras: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Feedback:
    """Feedback data structure."""
    id: tuple  # (agentId, clientAddress, feedbackIndex) - tuple for efficiency
    agentId: AgentId
    reviewer: Address
    score: Optional[int]  # 0-100
    tags: List[str] = field(default_factory=list)
    text: Optional[str] = None
    context: Optional[Dict[str, Any]] = None
    proofOfPayment: Optional[Dict[str, Any]] = None
    fileURI: Optional[URI] = None
    createdAt: Timestamp = field(default_factory=lambda: int(datetime.now().timestamp()))
    answers: List[Dict[str, Any]] = field(default_factory=list)
    isRevoked: bool = False
    
    # Off-chain only fields (not stored on blockchain)
    capability: Optional[str] = None  # MCP capability: "prompts", "resources", "tools", "completions"
    name: Optional[str] = None  # MCP tool/resource name
    skill: Optional[str] = None  # A2A skill
    task: Optional[str] = None  # A2A task
    
    def __post_init__(self):
        """Validate and set ID after initialization."""
        if isinstance(self.id, str):
            # Convert string ID to tuple
            parsed_id = self.from_id_string(self.id)
            self.id = parsed_id
        elif not isinstance(self.id, tuple) or len(self.id) != 3:
            raise ValueError(f"Feedback ID must be tuple of (agentId, clientAddress, feedbackIndex), got: {self.id}")
    
    @property
    def id_string(self) -> str:
        """Get string representation of ID for external APIs."""
        return f"{self.id[0]}:{self.id[1]}:{self.id[2]}"
    
    @classmethod
    def create_id(cls, agentId: AgentId, clientAddress: Address, feedbackIndex: int) -> tuple:
        """Create feedback ID tuple with normalized address."""
        # Normalize address to lowercase for consistency
        # Ethereum addresses are case-insensitive, but we store them in lowercase
        if isinstance(clientAddress, str) and (clientAddress.startswith("0x") or clientAddress.startswith("0X")):
            normalized_address = "0x" + clientAddress[2:].lower()
        else:
            normalized_address = clientAddress.lower() if isinstance(clientAddress, str) else str(clientAddress).lower()
        return (agentId, normalized_address, feedbackIndex)
    
    @classmethod
    def from_id_string(cls, id_string: str) -> tuple:
        """Parse feedback ID from string.
        
        Format: agentId:clientAddress:feedbackIndex
        Note: agentId may contain colons (e.g., "11155111:123"), so we need to split from the right.
        """
        parts = id_string.rsplit(":", 2)  # Split from right, max 2 splits
        if len(parts) != 3:
            raise ValueError(f"Invalid feedback ID format: {id_string}")
        
        try:
            feedback_index = int(parts[2])
        except ValueError:
            raise ValueError(f"Invalid feedback index: {parts[2]}")
        
        # Normalize address to lowercase for consistency
        client_address = parts[1]
        if isinstance(client_address, str) and (client_address.startswith("0x") or client_address.startswith("0X")):
            normalized_address = "0x" + client_address[2:].lower()
        else:
            normalized_address = client_address.lower() if isinstance(client_address, str) else str(client_address).lower()
        
        return (parts[0], normalized_address, feedback_index)


@dataclass
class SearchParams:
    """Parameters for agent search."""
    chains: Optional[Union[List[ChainId], Literal["all"]]] = None
    name: Optional[str] = None  # case-insensitive substring
    description: Optional[str] = None  # semantic; vector distance < threshold
    owners: Optional[List[Address]] = None
    operators: Optional[List[Address]] = None
    mcp: Optional[bool] = None
    a2a: Optional[bool] = None
    ens: Optional[str] = None  # exact, case-insensitive
    did: Optional[str] = None  # exact
    walletAddress: Optional[Address] = None
    supportedTrust: Optional[List[str]] = None
    a2aSkills: Optional[List[str]] = None
    mcpTools: Optional[List[str]] = None
    mcpPrompts: Optional[List[str]] = None
    mcpResources: Optional[List[str]] = None
    active: Optional[bool] = True
    x402support: Optional[bool] = None
    deduplicate_cross_chain: bool = False  # Deduplicate same agent across chains

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary, filtering out None values."""
        return {k: v for k, v in self.__dict__.items() if v is not None}


@dataclass
class SearchFeedbackParams:
    """Parameters for feedback search."""
    agents: Optional[List[AgentId]] = None
    tags: Optional[List[str]] = None
    reviewers: Optional[List[Address]] = None
    capabilities: Optional[List[str]] = None
    skills: Optional[List[str]] = None
    tasks: Optional[List[str]] = None
    names: Optional[List[str]] = None  # MCP tool/resource/prompt names
    minScore: Optional[int] = None  # 0-100
    maxScore: Optional[int] = None  # 0-100
    includeRevoked: bool = False
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary, filtering out None values."""
        return {k: v for k, v in self.__dict__.items() if v is not None}
